import logging
from functools import lru_cache
from typing import Any, Callable, Dict, Iterable, List, Optional, Tuple, Union

import pandas as pd

from elmada import cc_share, from_entsoe, from_geo_scraped, from_opsd, from_other
from elmada import mappings as mp

logger = logging.getLogger(__name__)
logger.setLevel(level=logging.WARN)


def prep_prices(year=2019, freq="60min", country="DE", **kwargs) -> pd.Series:
    return prep_CEFs(year=year, freq=freq, country=country, **kwargs)["marginal_cost"]


def prep_CEFs(
    year: int = 2019,
    freq: str = "60min",
    country: str = "DE",
    validation_mode: bool = False,
    mo_P: Optional[pd.DataFrame] = None,
    **mo_kwargs,
) -> pd.DataFrame:
    """Prepares XEFs for European countries with piece-wise-linear approximation method"""

    if mo_P is None:
        mo_P = merit_order(year=year, country=country, validation_mode=validation_mode, **mo_kwargs)
    return from_opsd.get_CEFs_from_merit_order(mo_P=mo_P, year=year, freq=freq, country=country)


def merit_order(
    year: int = 2019,
    country: str = "DE",
    approx_method: str = "regr",
    validation_mode: bool = False,
    overwrite_carbon_tax: Optional[float] = None,
    pp_size_method: str = "from_geo_scraped",
) -> pd.DataFrame:
    """Return a merit-order list. Virtual power plants are constructed through discretization."""

    mo_f = merit_order_per_fuel(
        year=year, country=country, approx_method=approx_method, validation_mode=validation_mode,
    )

    df = discretize_merit_order_per_fuel(mo_f, pp_size_method=pp_size_method, country=country)

    df["x_k"] = df["fuel_draf"].map(from_other.get_fuel_prices(year=year, country=country))

    DATA_QUASCH = from_other.get_emissions_per_fuel_quaschning()
    marginal_emissions_of_gen = df["fuel_draf"].map(DATA_QUASCH) / df["used_eff"]
    df["fuel_cost"] = df["x_k"] / df["used_eff"]

    carbon_price = (
        from_other.get_ETS_price(year) if overwrite_carbon_tax is None else overwrite_carbon_tax
    )
    df["GHG_cost"] = marginal_emissions_of_gen * carbon_price
    df["marginal_emissions"] = marginal_emissions_of_gen
    df["marginal_cost"] = df["fuel_cost"] + df["GHG_cost"]

    df = df.sort_values("marginal_cost").reset_index(drop=True)
    df["cumsum_capa"] = df["capa"].cumsum()
    df = df.dropna()
    return df.copy()


def merit_order_per_fuel(
    year: int = 2019,
    country: str = "DE",
    approx_method: str = "regr",
    validation_mode: bool = False,
) -> pd.DataFrame:
    """Returns a non-discretized merit order on fuel-type level."""
    eff_min, eff_max, __, __ = approximate_min_max_values(method=approx_method)

    source = "power_plant_list" if validation_mode else "entsoe"

    df = pd.DataFrame(
        dict(
            eff_min=eff_min,
            eff_max=eff_max,
            fuel_price=from_other.get_fuel_prices(year=year, country=country),
            emissions_for_gen=from_other.get_emissions_per_fuel_quaschning(),  # in t_CO2eq / MWh_el
            capa=prep_installed_generation_capacity(year=year, country=country, source=source),
        )
    )

    logger.info(f"Available fuels for {year}, {country}: {list(df.index)}")

    df["eff_mean"] = (df["eff_min"] + df["eff_max"]) / 2
    df["fuel_cost"] = df["fuel_price"] / df["eff_mean"]
    df["GHG_cost"] = df["emissions_for_gen"] / df["eff_mean"]
    df["marginal_cost"] = df["fuel_cost"] + df["GHG_cost"]
    df = df.sort_values("marginal_cost")
    df["cumsum_capa"] = df["capa"].cumsum()
    return df.copy()


@lru_cache(maxsize=2)
def approximate_min_max_values(method="regr") -> Tuple[Dict, Dict, Dict, Dict]:
    """`method` must be either 'interp' for interpolation or 'regr' for linear regression."""
    mo = get_clean_merit_order_for_min_max_approximation(sort_by_fuel=True)

    if method == "interp":
        grouper = mo.groupby("fuel_draf")["used_eff"]
        eff_min = grouper.min().to_dict()
        eff_max = grouper.max().to_dict()

        grouper = mo.groupby("fuel_draf")["cumsum_capa"]
        cumsum_capa_LHS = grouper.min().to_dict()
        cumsum_capa_RHS = grouper.max().to_dict()

    elif method == "regr":
        from scipy import stats

        eff_min = {}
        eff_max = {}
        cumsum_capa_LHS = {}
        cumsum_capa_RHS = {}

        for fuel in mo["fuel_draf"].unique():
            mo_ = mo[mo["fuel_draf"] == fuel]
            slope, intercept, r_value, p_value, std_err = stats.linregress(
                mo_["cumsum_capa"], mo_["used_eff"]
            )

            cumsum_capa_LHS[fuel] = mo_["cumsum_capa"].min()
            cumsum_capa_RHS[fuel] = mo_["cumsum_capa"].max()

            eff_min[fuel] = intercept + slope * cumsum_capa_RHS[fuel]
            eff_max[fuel] = intercept + slope * cumsum_capa_LHS[fuel]

    else:
        raise ValueError(f"Method must be either 'interp' or 'regr'. Given:'{method}'")

    return eff_min, eff_max, cumsum_capa_RHS, cumsum_capa_LHS


def get_clean_merit_order_for_min_max_approximation(sort_by_fuel=False) -> pd.DataFrame:
    """Return a clean German merit order of the generation technologies."""

    # The opsd-file is used here here!
    mo = from_opsd.merit_order(year=2019)

    if sort_by_fuel:
        groups = []
        for fuel, group in mo.groupby("fuel_draf", sort=False):
            groups.append(group.sort_values(by="used_eff", ascending=False))

        mo = pd.concat(groups, axis=0, ignore_index=True)

        mo["cumsum_capa"] = mo["capa"].cumsum()

    return mo


@lru_cache(maxsize=3)
def prep_installed_generation_capacity(year=2019, country="DE", source="entsoe") -> pd.Series:

    if source == "power_plant_list":
        ser = (
            from_opsd.merit_order(year=year)[["fuel_draf", "capa"]]
            .groupby("fuel_draf")
            .sum()["capa"]
        )
        return ser

    elif source == "entsoe":
        df = from_entsoe.load_installed_generation_capacity(year=year, country=country)
        df = from_entsoe.aggregate_to_standard_techs(df)
        ser = df.T.iloc[:, 0]

        ser = apply_share_cc_assumption(ser, country=country)

        logger.info(f"Available fuels for {year}, {country}: {list(ser.index)}")

        ser = pd.Series(index=mp.PWL_FUELS, data=ser)
        ser.fillna(0, inplace=True)
        return ser

    # elif source == "opsd":
    #     ser = from_opsd.get_installed_generation_capacity(year=year, country=country)
    #     ser = pd.Series(index=mp.PWL_FUELS, data=ser)
    #     ser.fillna(0, inplace=True)
    #     ser = apply_share_cc_assumption(ser, country=country)
    #     return ser

    else:
        raise ValueError("Source must be either 'entsoe' or 'power_plant_list'")


def apply_share_cc_assumption(ser, country: str) -> pd.Series:
    if "gas" in ser:
        share_cc = get_share_cc(country=country)
        ser["gas_cc"] = ser["gas"] * share_cc
        ser["gas"] = ser["gas"] * (1 - share_cc)

    return ser


def get_share_cc(country: str) -> float:
    try:
        return cc_share.get_ccgt_shares_from_cascade().loc[country, "selected"]
    except KeyError:
        logger.warning("Unable to apply cascade method, German values returned")
        return cc_share.get_ccgt_DE()


def discretize_merit_order_per_fuel(
    mo_f: pd.DataFrame, country: str, pp_size_method: str = "from_geo_scraped"
) -> pd.DataFrame:

    concat_list = []
    # if country=="DE":
    #     pp_size_method = "from_Germany"
    for fuel, row in mo_f.iterrows():
        pp_size = get_pp_size(pp_size_method=pp_size_method, country=country, fuel=fuel)
        logger.info(f"pp_size_method={pp_size_method}")
        number_of_powerplants = int(row["capa"] // pp_size)
        capa_of_last_powerplant = row["capa"] % pp_size
        df = pd.DataFrame({"capa": [pp_size] * number_of_powerplants + [capa_of_last_powerplant]})
        cumsum_capa_in_f = df["capa"].cumsum()
        df["used_eff"] = row["eff_max"] - (cumsum_capa_in_f / row["capa"]) * (
            row["eff_max"] - row["eff_min"]
        )
        df["fuel_draf"] = fuel
        concat_list.append(df)
    df = pd.concat(concat_list, ignore_index=True)
    return df


def get_pp_size(pp_size_method: str, country: str, fuel: str) -> int:
    try:
        if pp_size_method == "from_geo_scraped":
            return from_geo_scraped.get_pp_sizes_for_pwl().loc[country, fuel]
        # elif pp_size_method == "from_Germany":
        #     return get_pp_sizes_from_germany()[fuel]
        # elif pp_size_method == "from_geo":
        #     return from_geo_via_morph.get_pp_sizes().loc[country, fuel]
        # elif pp_size_method == "from_ccgt":
        #     return cc_share.get_pp_sizes().loc[country, fuel]
        else:
            raise ValueError(f"Invalid pp_size_method given: {pp_size_method}")
    except KeyError:
        logger.warning("Unable to apply pp-size, German values returned")
        return get_pp_sizes_from_germany()[fuel]


@lru_cache(maxsize=1)
def get_pp_sizes_from_germany() -> pd.Series:
    return from_opsd.merit_order().groupby("fuel_draf").capa.mean()



import hashlib
from pathlib import Path

import pandas as pd
from numpy.core.arrayprint import str_format

import elmada
from elmada import mappings as mp

HASH_DIR = Path(__file__).resolve().parent
FILE_FMT = "{which}_hashes.csv"


def read_expected_hashes(which: str) -> pd.Series:
    fp = HASH_DIR / FILE_FMT.format(which=which)
    return pd.read_csv(fp, index_col=[0, 1], squeeze=True)


def save_expected_hashes(which: str) -> None:
    fp = HASH_DIR / FILE_FMT.format(which=which)
    if input("Do you really want to overwrite the data hash table? (y/n)") == "y":
        make_hashes(which=which).to_csv(fp, header=True)
    else:
        print("Aborted.")


def make_hashes(which) -> pd.Series:

    if which == "CEF":
        func = elmada.get_emissions
        kwargs = dict(freq="60min", method="_PWL", cache=True)

    elif which == "GEN":
        func = elmada.get_el_national_generation
        kwargs = dict(freq="60min")

    else:
        raise ValueError("`which` must be 'CEF' or 'GEN'.")

    return make_year_country_hashes(func=func, kwargs=kwargs)


def make_year_country_hashes(func, kwargs) -> pd.Series:
    ser = pd.Series(
        {
            (year, country): get_hash(func(year=year, country=country, **kwargs))
            for year in range(2017, 2021)
            for country in mp.COUNTRIES_FOR_ANALYSIS
        },
        name="hash",
    )
    ser.index.names = ["year", "country"]
    return ser


def get_hash(df) -> str_format:
    return hashlib.sha256(pd.util.hash_pandas_object(df, index=True).values).hexdigest()[:10]



from pathlib import Path

import pandas as pd

import elmada
from elmada import mappings as mp

MYDIR = Path(__file__).resolve().parent


def make_avgs() -> pd.DataFrame:
    return pd.Series(
        {
            (str(year), country, cef_type): int(
                elmada.get_emissions(year=year, country=country, method=f"{cef_type}_PWL").mean()
            )
            for year in range(2017, 2021)
            for country in mp.COUNTRIES_FOR_ANALYSIS
            for cef_type in ("MEF", "XEF")
        },
    ).unstack([0])


def save_expected_avgs() -> None:
    fp = MYDIR / "avgs.csv"
    if input("Do you really want to overwrite the data average table? (y/n)") == "y":
        make_avgs().to_csv(fp, header=True)
    else:
        print("Aborted.")


def read_expected_avgs() -> pd.Series:
    fp = MYDIR / "avgs.csv"
    return pd.read_csv(fp, index_col=[0, 1])



import logging
import warnings

import pandas as pd

from elmada import helper as hp
from elmada import paths

logger = logging.getLogger(__name__)
logger.setLevel(level=logging.WARN)




def prep_dayahead_prices(
    year: int, freq: str = "60min", country: str = "DE", cache: bool = True
) -> pd.Series:

    assert year in range(2000, 2100), f"{year} is not a valid year"
    assert country == "DE", "Smard is only used for Germany!"

    fp = paths.CACHE_DIR / f"{year}_{freq}_{country}_dayahead_c_el_smard.parquet"

    if cache and fp.exists():
        ser = hp.read(fp)

    else:
        fp_raw = paths.DATA_DIR / f"smard/Day-ahead_prices_{year}.csv"
        df = pd.read_csv(
            fp_raw,
            sep=";",
            usecols=["Germany/Luxembourg[€/MWh]", "Germany/Austria/Luxembourg[€/MWh]"],
            na_values="-",
        )
        df = df.reset_index(drop=True)
        df = df.fillna(df.mean())

        # add a merged column
        df["DE_merged"] = 0

        if year < 2018:
            df["DE_merged"] = df["Germany/Austria/Luxembourg[€/MWh]"]

        elif year == 2018:

            warnings.filterwarnings(
                "ignore",
                message="A value is trying to be set on a copy of a slice from a DataFrame",
            )
            df.loc[:6552, "DE_merged"] = df.loc[:6552, "Germany/Austria/Luxembourg[€/MWh]"]
            df.loc[6552:, "DE_merged"] = df.loc[6552:, "Germany/Luxembourg[€/MWh]"]

        elif year > 2018:
            df["DE_merged"] = df["Germany/Luxembourg[€/MWh]"]

        ser = df["DE_merged"]
        if cache:
            hp.write(ser, fp)

    start_freq = hp.estimate_freq(ser)

    ser = hp.resample(ser, year=year, start_freq=start_freq, target_freq=freq)
    if start_freq != freq:
        logger.warning(f"Dayahead_prices were resampled from {start_freq} to {freq}")

    hp.warn_if_incorrect_index_length(ser, year, freq)
    return ser



import pandas as pd
import requests
from bs4 import BeautifulSoup

from elmada import from_geo_via_morph
from elmada import helper as hp
from elmada import mappings as mp
from elmada import paths


def get_pp_sizes_for_pwl() -> pd.DataFrame:
    df = get_pp_sizes()
    df = df.fillna(df.mean()).astype(int)
    return df


def get_pp_sizes() -> pd.DataFrame:
    geo = get_units_of_geo_list()

    grouper = geo.groupby(["cy", "fuel"])["capa"]
    pp_sizes = grouper.mean().unstack()
    count = grouper.count().unstack()
    pp_sizes = pp_sizes.where(count >= 2)
    return pp_sizes[mp.PWL_FUELS]


def get_units_of_geo_list(cache: bool = True) -> pd.DataFrame:
    fp = paths.mode_dependent_cache_dir() / "units_of_geo_list.parquet"

    if fp.exists() and cache:
        df = hp.read(fp)

    else:
        df = _query_geo_power_plant_data()
        if cache:
            hp.write(df, fp)

    df = df.rename(
        columns={
            "Capacity (MWe)": "capa",
            "Unit Efficiency (%)": "eff",
            "Date Commissioned (yyyy-mm-dd)": "commissioned",
            "Unit #": "unit_no",
        },
    )

    interesting_cols = ["cy", "fuel", "geoid", "capa", "eff", "commissioned", "unit_no"]
    df = df.loc[df.capa != "", interesting_cols]

    df["capa"] = df["capa"].astype(float).astype(int)
    return df.reset_index(drop=True)


def _query_geo_power_plant_data():
    geo = from_geo_via_morph.get_geo_list()
    max_count = len(geo)
    print(f"Download {max_count} items:", end="")
    concat_list = []
    for _, ser in geo.iterrows():
        print(".", end="")
        geo_id = ser["id"]
        print(geo_id, end=", ")
        df = get_df_from_geo_id(geo_id)
        df["cy"] = ser["cy"]
        df["fuel"] = ser["fuel"]
        df["geoid"] = geo_id
        concat_list.append(df)
    df = pd.concat(concat_list)
    return df


def get_df_from_geo_id(geo_id: int) -> pd.DataFrame:
    url = f"http://globalenergyobservatory.org/geoid/{geo_id}"
    page = requests.get(url).text
    soup = BeautifulSoup(page, "lxml")
    selector = soup.find("div", {"id": "UnitDescription_Block"})

    table = selector.find("table")

    headers = [cell.get_text().strip() for cell in table.findAll("th")]
    ncols = len(headers)
    rows = table.findAll("tr")

    df = pd.DataFrame(columns=headers)
    for i, row in enumerate(rows):
        cells = row.find_all("td")
        if len(cells) == ncols:
            df.loc[i, :] = [cell.find("input").get("value") for cell in cells]
    return df



from pathlib import Path
from typing import Optional

from appdirs import user_cache_dir, user_config_dir

from elmada.mode import is_safe_mode


def _get_cache_directory() -> Path:
    """Returns the path to cache directory and creates it, if not yet existing."""
    fp = Path(user_cache_dir(appname="Elmada", appauthor="DrafProject"))
    fp.mkdir(parents=True, exist_ok=True)
    return fp


def _get_config_directory() -> Path:
    """Returns the path to config directory and creates it, if not yet existing."""
    fp = Path(user_config_dir(appname="Elmada", appauthor="DrafProject")) / "api_keys"
    fp.mkdir(parents=True, exist_ok=True)
    return fp


BASE_DIR = Path(__file__).resolve().parent
CACHE_DIR = _get_cache_directory()
KEYS_DIR = _get_config_directory()
DATA_DIR = BASE_DIR / "data/raw"
SAFE_CACHE_DIR = BASE_DIR / "data/safe_cache"


def mode_dependent_cache_dir(year: Optional[int] = None):
    is_safe_year = year in range(2017, 2021) or year is None
    return SAFE_CACHE_DIR if (is_safe_mode() and is_safe_year) else CACHE_DIR



"""Handles power plant list from Global Energy Observatory via morph."""

import collections
import logging
import sqlite3
from functools import lru_cache
from pathlib import Path
from typing import Any, Callable, Dict, Iterable, List, Optional, Set, Tuple, Union

import numpy as np
import pandas as pd
import requests

from elmada import helper as hp
from elmada import mappings as mp
from elmada import paths

logger = logging.getLogger(__name__)
logger.setLevel(logging.WARN)

DB_FILE_NAME = Path("GEO-database.suffix")

CC_WEIGHTING = {
    "Thermal and CCGT": 1,
    "Power and Heat Combined Cycle Gas Turbine": 1,
    "Combined Cycle Gas Engine (CCGE)": 1,
    "Combined Cycle Gas Turbine": 1,
    "OCGT and CCGT": 0.49,
    "Sub-critical Thermal": 0,
    "Power and Heat Open Cycle Gas Turbine": 0,
    "Open Cycle Gas Turbine": 0,
    "Heat and Power Steam Turbine": 0,
    "Super-critical Thermal": 0,
    "Gas Engines": 0,
    None: 0,
}

COLS = {
    "Name": "name",
    "Type": "fuel",
    "Country": "country",
    "Type_of_Plant_rng1": "plant_type",
    "Type_of_Fuel_rng1_Primary": "primary_fuel",
    "Design_Capacity_MWe_nbr": "capa",
    "GEO_Assigned_Identification_Number": "id",
    "Status_of_Plant_itf": "status",
}

NON_OPERATIONAL_STATUSES = [
    "Built and In Test Stage",
    "Decommissioned",
    "Demolished",
    "Design and Planning Stage",
    "Final Bid and Approval Stage",
    "Mothballed Full",
    "Mothballed Partial",
    "Shutdown" "Under Construction",
]


def get_ccgt_shares() -> pd.DataFrame:
    geo = get_geo_list()
    valid_countries = get_valid_countries()
    df = pd.DataFrame(index=valid_countries.keys())

    for cy_short, cy_long in valid_countries.items():
        is_gas = geo["fuel"].isin(["gas", "gas_cc"])
        is_country = geo["country"] == cy_long
        geo_sub = geo[is_gas & is_country]
        df.loc[cy_short, "cy_long"] = cy_long
        df.loc[cy_short, "cc_capa"] = geo_sub["capa"] @ geo_sub["cc_weight"]
        df.loc[cy_short, "total_capa"] = geo_sub["capa"].sum()

    df["share_cc"] = df["cc_capa"] / df["total_capa"]
    return df


def get_valid_countries():
    df = get_geo_list()
    avail = set(df["country"].unique())
    d = {short: long for short, long in mp.COUNTRIES_FOR_ANALYSIS.items() if long in avail}
    return collections.OrderedDict(sorted(d.items()))


@lru_cache(maxsize=1)
def get_geo_list():
    df = get_geo_full_list()
    df = filter_relevant_plants(df)
    df = filter_operating_plants(df)
    df = set_lignite_plants(df)
    df = set_gasCC_plants(df)

    df["fuel"] = df["fuel"].str.lower()

    reverse_ana = {v: k for k, v in mp.COUNTRIES_FOR_ANALYSIS.items()}
    df.insert(loc=3, column="cy", value=df["country"].map(reverse_ana))

    return df[df.capa.notnull()]


def set_gasCC_plants(df: pd.DataFrame) -> pd.DataFrame:
    is_gas = df["fuel"] == "Gas"

    df["cc_weight"] = np.nan
    df.loc[is_gas, "cc_weight"] = df.loc[is_gas, "plant_type"].map(CC_WEIGHTING)
    # is_cc = df["cc_weight"][is_gas].apply(round).astype("bool")
    is_cc = df["cc_weight"] >= 0.5
    df.loc[is_gas & is_cc, "fuel"] = "Gas_cc"
    return df


def set_lignite_plants(df: pd.DataFrame) -> pd.DataFrame:
    is_coal = df["fuel"] == "Coal"
    is_lignite = df["primary_fuel"].str.contains("lignite", case=False)
    is_peat = df["primary_fuel"].str.contains("peat", case=False)
    df.loc[is_coal & (is_lignite ^ is_peat), "fuel"] = "Lignite"
    return df


def filter_operating_plants(df: pd.DataFrame) -> pd.DataFrame:
    no_status = df.status.isna()
    contains_shutdown = df.name.str.contains("(Shutdown)")
    df.loc[no_status & contains_shutdown, "status"] = "Shutdown"
    is_operating = ~df["status"].isin(NON_OPERATIONAL_STATUSES)
    return df[is_operating]


def filter_relevant_plants(df: pd.DataFrame) -> pd.DataFrame:
    is_ana = df["Country"].isin(mp.COUNTRIES_FOR_ANALYSIS.values())
    is_conv = df["Type"].isin(["Gas", "Coal", "Nuclear", "Oil"])
    return df.loc[is_ana & is_conv, COLS.keys()].rename(columns=COLS)


def get_geo_full_list(cache: bool = True) -> pd.DataFrame:

    fp = paths.mode_dependent_cache_dir() / DB_FILE_NAME.with_suffix(".parquet")

    if cache and fp.exists():
        df = hp.read(fp)

    else:
        fp_db = paths.CACHE_DIR / DB_FILE_NAME.with_suffix(".db")

        if not fp_db.exists():
            download_database(fp=fp_db)

        conn = sqlite3.connect(fp_db)
        df = pd.read_sql_query("SELECT * FROM powerplants", conn)
        conn.close()
        if cache:
            hp.write(df, fp)

    return df


def download_database(fp: Path) -> None:
    url_base = "https://morph.io/coroa/global_energy_observatory_power_plants/"
    morph_api_key = hp.get_api_key("morph")
    url_ending = f"data.sqlite?key={morph_api_key}"
    url = url_base + url_ending

    response = requests.get(url)
    logger.info(f"{fp.name} downloaded from GEO via Morph.")
    fp.write_bytes(response.content)



"""Web-scrapers for the CCGT-share of the gas-fired power plants"""
import collections
import logging
from functools import lru_cache
from typing import Any, Callable, Dict, Iterable, List, Optional, Set, Tuple, Union

import numpy as np
import pandas as pd
import requests
from bs4 import BeautifulSoup

from elmada import eu_pwl, from_geo_via_morph
from elmada import helper as hp
from elmada import mappings as mp
from elmada import paths

logger = logging.getLogger(__name__)
logger.setLevel(logging.WARN)


@lru_cache(maxsize=1)
def get_ccgt_shares_from_cascade():
    rich_geo = from_geo_via_morph.get_ccgt_shares()["share_cc"].rename("rich_geo")
    wiki = get_wiki_shares(cache=True).rename("manual")
    opsd = pd.Series({"DE": get_ccgt_DE()}).rename("opsd")

    df = pd.concat([wiki, rich_geo, opsd], axis=1, sort=False)

    def select_source(row):
        if not np.isnan(row["opsd"]):
            return "opsd"
        elif not np.isnan(row["rich_geo"]):
            return "rich_geo"
        elif not np.isnan(row["manual"]):
            return "manual"

    def get_selected(row):
        source = row["source"]
        return row[source]

    df["source"] = df.apply(select_source, 1)
    df["selected"] = df.apply(get_selected, 1)
    # df.reindex(mp.COUNTRIES_FOR_ANALYSIS.keys())
    df = df.sort_index()
    return df


#################################
# OPSD POWER PLANT LIST         #
#################################


def get_ccgt_DE() -> pd.Series:
    ser = eu_pwl.prep_installed_generation_capacity(
        year=2019, country="DE", source="power_plant_list"
    )
    return ser["gas_cc"] / (ser["gas_cc"] + ser["gas"])


#################################
# GEO LIST self-scraped         #
#################################


def get_geo_shares_overview() -> pd.DataFrame:
    geo = get_geo_list()
    valid_countries = get_valid_countries()
    header = ["cy_long", "nplants", "total_capa"]
    df = pd.DataFrame(index=valid_countries.keys(), columns=header)

    for cy_short, cy_long in valid_countries.items():
        is_country = geo.country == cy_long
        df.loc[cy_short, "cy_long"] = cy_long
        df.loc[cy_short, "nplants"] = is_country.sum()
        df.loc[cy_short, "cc_capa"] = geo.loc[is_country & geo.is_ccgt, "capa"].sum()
        df.loc[cy_short, "total_capa"] = geo.loc[is_country, "capa"].sum()

    df["share_cc"] = df["cc_capa"] / df["total_capa"]
    return df


def get_valid_countries() -> collections.OrderedDict:
    df = get_geo_list()
    avail = set(df["country"].unique())
    d = {short: long for short, long in mp.COUNTRIES_FOR_ANALYSIS.items() if long in avail}
    return collections.OrderedDict(sorted(d.items()))


def get_geo_list(
    country_long: Optional[str] = None,
    filter: Optional[str] = None,
    fuel="gas",
    cache: bool = True,
) -> pd.DataFrame:
    """possible conventional fuels are: gas, coal, nuclear, oil"""
    fp = paths.mode_dependent_cache_dir() / f"geo_list_{fuel}.parquet"

    if cache and fp.exists():
        df = hp.read(fp)

    else:
        fuel_ = fuel.capitalize()
        base_url = "http://globalenergyobservatory.org/"
        url = base_url + f"list.php?db=PowerPlants&type={fuel_}"
        page = requests.get(url).text
        soup = BeautifulSoup(page, "lxml")
        my_table = soup.find("table")
        headers = [cell.get_text().strip() for cell in my_table.find_all("h2")]
        ncols = len(headers)
        entries = my_table.findAll("td")

        df = pd.DataFrame(
            {
                "name": [x.get_text() for x in entries[ncols + 0 :: 4]],
                "geoid": [
                    x.find("a").get_attribute_list("href")[0].split("/")[1].strip()
                    for x in entries[ncols + 0 :: 4]
                ],
                "capa": [cell.get_text().strip() for cell in entries[ncols + 1 :: 4]],
                "country": [cell.get_text().strip() for cell in entries[ncols + 2 :: 4]],
                "state": [cell.get_text().strip() for cell in entries[ncols + 3 :: 4]],
            }
        )
        # df.loc[df["country"] == "Czech Republic", "country"] = "Czechia"
        df.loc[df.capa == ""] = np.nan
        df["is_ccgt"] = df.name.str.contains("CCGT").astype(bool)
        df["capa"] = df["capa"].astype(np.float)
        if cache:
            hp.write(df, fp)

    if country_long is None and filter is None:
        return df
    elif country_long is not None and filter is None:
        return df[df["country"] == country_long]
    elif country_long is None and filter is not None:
        if filter == "ana":
            return df[df.country.isin(mp.COUNTRIES_FOR_ANALYSIS.values())]
        elif filter == "eur":
            return df[df.country.isin(mp.EUROPE_COUNTRIES.values())]
        else:
            ValueError("`filter` must be either 'ana' or 'eur'.")
    else:
        raise ValueError("Country_long and filter given but only one of the two expected.")


#################################
# MANUAL                        #
#################################


def get_wiki_shares(cache: bool = True) -> pd.Series:
    fp = paths.mode_dependent_cache_dir() / "share_ccgt_wiki.parquet"
    if cache and fp.exists():
        ser = hp.read(fp)
    else:
        shares = {
            "AT": get_ccgt_AT(),
            "DK": get_ccgt_DK(),
            "IE": get_ccgt_IE(),
            "IT": get_ccgt_IT(),
            "RS": get_ccgt_RS(),
        }
        ser = pd.Series(shares)
        if cache:
            hp.write(ser, fp)
    return ser


def get_ccgt_IE():
    url = "https://en.wikipedia.org/w/index.php?&oldid=942359418"
    page = requests.get(url).text
    soup = BeautifulSoup(page, "lxml")

    my_table = soup.find("table", {"class": "wikitable sortable"})
    headers = [cell.get_text() for cell in my_table.findAll("th")]
    ncols = len(headers)
    rows = my_table.findAll("tr")

    df = pd.DataFrame(columns=headers)
    for i, row in enumerate(rows):
        cells = row.find_all("td")
        if len(cells) == ncols:
            df.loc[i, :] = [cell.get_text() for cell in cells]

    is_gas = df["Primary Fuel"] == "Gas"
    is_cc = df["Cycle"] == "Combined Cycle"
    capa_gasCC = df.loc[is_gas & is_cc, "Capacity (MW)"].astype("int").sum()
    capa_gas = df.loc[is_gas, "Capacity (MW)"].astype("int").sum()
    return capa_gasCC / capa_gas


def get_ccgt_AT():
    url = "https://de.wikipedia.org/w/index.php?oldid=199043393"
    page = requests.get(url).text
    soup = BeautifulSoup(page, "lxml")

    my_table = soup.find_all("table", {"class": "wikitable sortable"})[3]

    headers = [cell.get_text().rstrip("\n") for cell in my_table.findAll("th")]
    ncols = len(headers)
    rows = my_table.findAll("tr")

    df = pd.DataFrame(columns=headers)
    for i, row in enumerate(rows):
        cells = row.find_all("td")
        if len(cells) == ncols:
            df.loc[i, :] = [cell.get_text().rstrip("\n") for cell in cells]

    df = (
        df.replace(r"\[(.*?)\]", "", regex=True)
        .replace(r"\n", "", regex=True)
        .replace(",", ".", regex=True)
    )
    df.columns = [col.replace("\n", "") for col in df.columns]
    cap_col = "Leistung Elektrischin MWel"
    df[cap_col] = df[cap_col].astype("float")

    # manually entered by research on de.wikipedia.org pages on 2020-04-24:
    is_ccgt = {
        "Gas- und Dampfkraftwerk Mellach": True,
        "Kraftwerk Theiß": True,
        "Kraftwerk Simmering \xa01": False,
        "Kraftwerk Simmering \xa03": False,
        "Kraftwerk Simmering \xa02": True,
        "Gas- und Dampfkraftwerk Timelkam": False,  # n/a
        "Dampfkraftwerk Donaustadt \xa03": True,
        "Fernheizkraftwerk Linz-Mitte": True,
        "Linz Süd": False,  # n/a
        "Kraftwerk Korneuburg \xa01": True,
        "Gas- und Dampfkraftwerk Leopoldau": True,
        "Heizkraftwerk Salzburg Mitte": False,
        "Heizkraftwerk Salzburg Nord": False,
        "Müllverbrennungsanlage Spittelau": False,
    }

    df["is_ccgt"] = df.Name.map(is_ccgt)
    return df[df.is_ccgt][cap_col].sum() / df[cap_col].sum()


def get_ccgt_IT():
    url = "https://de.wikipedia.org/w/index.php?&oldid=194656154"
    page = requests.get(url).text
    soup = BeautifulSoup(page, "lxml")

    my_table = soup.find_all("table", {"class": "wikitable sortable"})[1]

    headers = [cell.get_text() for cell in my_table.findAll("th")]
    ncols = len(headers)
    rows = my_table.findAll("tr")

    df = pd.DataFrame(columns=headers)
    for i, row in enumerate(rows):
        cells = row.find_all("td")
        if len(cells) == ncols:
            df.loc[i, :] = [cell.get_text() for cell in cells]

    df = df.replace(r"\[(.*?)\]", "", regex=True).replace(r"\n", "", regex=True)
    df.columns = [col.replace("\n", "") for col in df.columns]

    cap_col = "Inst. Leistung (MW)"
    df[cap_col] = df[cap_col].astype("float")
    df = df[(df.Brennstoff == "Erdgas") | (df.Brennstoff == "Erdgas/Kohle")].reset_index(drop=True)

    # manually entered by research on 2020-04-28:
    geo_url = "http://globalenergyobservatory.org/geoid/"
    is_ccgt = {
        "Alessandro Volta": [1, "https://de.wikipedia.org/w/index.php?oldid=194580946"],
        "Aprilia": [1, geo_url + "43990"],
        "Bertonico-Turano": [1, geo_url + "43994"],
        "Brindisi (A2A) (Brindisi Nord)": [1, geo_url + "43701"],
        "Brindisi (EniPower)": [1, geo_url + "43980"],
        "Cassano d’Adda": [1, geo_url + "2583"],
        "Chivasso": [1, geo_url + "43682"],
        "Ferrara": [1, geo_url + "43699"],
        "Ferrera Erbognone": [1, geo_url + "43700"],
        "Gissi": [1, geo_url + "43708"],
        "La Casella": [1, geo_url + "43771"],
        "La Spezia 'Eugenio Montale'": [1, geo_url + "43724"],
        "Livorno Ferraris": [1, geo_url + "43684"],
        "Ostiglia": [1, geo_url + "43694"],
        "Piacenza": [1, geo_url + "43692"],
        "Ravenna": [1, geo_url + "43702"],
        "Rossano": [0.5, geo_url + "44001"],
        "Scandale": [1, geo_url + "43696"],
        "Sermide": [1, geo_url + "43686"],
        "Simeri Crichi": [1, geo_url + "43979"],
        "Tavazzano": [1, geo_url + "43695"],
        "Torrevaldaliga Süd": [1, geo_url + "43713"],
        "Turbigo": [0.5, geo_url + "43685"],
        "Vado Ligure": [1, geo_url + "42714"],
    }

    df2 = pd.DataFrame.from_dict(is_ccgt, "index", columns=["is_ccgt", "source"]).reset_index(
        drop=True
    )
    df = pd.concat([df, df2], axis=1)
    df["is_ccgt"] = df["is_ccgt"].astype("float")

    return (df[cap_col] * df["is_ccgt"]).sum() / df[cap_col].sum()


def get_ccgt_DK():
    # manually entered by research on http://globalenergyobservatory.org/ pages on 2020-04-28:
    geo_url = "http://globalenergyobservatory.org/"
    df = pd.DataFrame(
        [
            ["HC Orsted CHP Power Plant Denmark", 185, True, geo_url + "form.php?pid=44235",],
            ["Skaerbaek CHP Power Plant Denmark", 392, False, geo_url + "form.php?pid=44237",],
            ["Svanemolle CHP Power Plant Denmark", 81, False, geo_url + "form.php?pid=44234",],
        ],
        columns=["name", "capa", "is_ccgt", "source"],
    )
    return df.loc[df["is_ccgt"], "capa"].sum() / df["capa"].sum()


def get_ccgt_RS():
    # 2020-04-29 manually entered by research wiki on https://en.wikipedia.org/w/index.php?oldid=944728190
    # : There is only one Gas-power plant in Serbia (Panonske TE-TO) which is not an CCGT.
    return 0.0



import logging
from typing import Any, Callable, Dict, Iterable, List, Optional, Set, Tuple, Union

import entsoe
import numpy as np
import pandas as pd
from entsoe import EntsoePandasClient
from entsoe import mappings as entsoe_mp

from elmada import exceptions
from elmada import helper as hp
from elmada import mappings as mp
from elmada import paths

logger = logging.getLogger(__name__)
logger.setLevel(level=logging.WARN)

# workaround for a bug in entsoe package v0.2.* see https://github.com/EnergieID/entsoe-py/issues/51
entsoe_mp.TIMEZONE_MAPPINGS["IT-NORD-AT"] = "Europe/Rome"
entsoe_mp.TIMEZONE_MAPPINGS["IT-NORD-FR"] = "Europe/Rome"
entsoe_mp.TIMEZONE_MAPPINGS["IT-GR"] = "Europe/Rome"
####################################################


def prep_XEFs(year: int = 2019, freq: str = "60min", country: str = "DE") -> pd.DataFrame:
    """Prepare grid mix emission factors from historic generation."""
    assert year in range(2000, 2100), f"{year} is not a valid year"
    assert freq in ["15min", "60min"], f"{freq} is not a valid freq"
    assert country in mp.EUROPE_COUNTRIES

    shares_TF = prep_shares(year, freq, country)
    ce_F = load_el_national_specific_emissions()[country]
    x = set(shares_TF.keys()) & set(ce_F.keys())
    ce_T = shares_TF[x] @ ce_F[x]
    ce_T = ce_T.fillna(ce_T.mean())

    hp.warn_if_incorrect_index_length(ce_T, year, freq)
    df = ce_T.rename("XEFs").to_frame()
    return df


def _get_client() -> EntsoePandasClient:
    return EntsoePandasClient(api_key=hp.get_api_key("entsoe"))


def _get_timestamps(year: int, tz: str) -> Tuple:
    start = pd.Timestamp(f"{year}0101", tz=tz)
    end = pd.Timestamp(f"{year+1}0101", tz=tz)
    return start, end


def _get_empty_installed_generation_capacity_df(ensure_std_techs: bool) -> pd.DataFrame:
    if ensure_std_techs:
        return pd.DataFrame(index=mp.DRAF_FUELS, columns=[""]).T
    else:
        raise exceptions.NoDataError("no Data")


def load_installed_generation_capacity(
    year: int = 2019, country: str = "DE", cache: bool = True, ensure_std_techs: bool = False,
) -> pd.DataFrame:
    """Returns a dataframe with installed generation capacity fuel type dependent on year and
    country.
    """

    fp = paths.mode_dependent_cache_dir() / f"{year}_{country}_installedGen_entsoe.parquet"
    warning = f"No installed generation capacity data available for {year}, {country}"

    if cache and fp.exists():
        df = hp.read(fp, squeeze=False)
    else:
        try:
            df = _query_installed_generation_capacity(year, country)

            if df.empty:
                logger.warning(f"{warning}: ENTSO-E returned an empty dataframe.")
                return _get_empty_installed_generation_capacity_df(ensure_std_techs)

            if cache:
                hp.write(df, fp)

        except (entsoe.exceptions.NoMatchingDataError, KeyError) as e:
            logger.error(f"{e}: {warning}")
            return _get_empty_installed_generation_capacity_df(ensure_std_techs)

    if ensure_std_techs:
        df = aggregate_to_standard_techs(df)

    return df


def _query_installed_generation_capacity(year, country) -> pd.DataFrame:
    client = _get_client()
    tz = get_timezone(country)
    start, end = _get_timestamps(year=year, tz=tz)
    df = client.query_installed_generation_capacity(
        start=start, end=end, country_code=country, psr_type=None
    )
    return df


def aggregate_to_standard_techs(df) -> pd.DataFrame:
    for i in ["other_conv", "other_RES"]:
        if i not in df:
            df[i] = 0

    df = df.rename(columns=mp.FUEL_RENAME)

    for k, v in mp.FUEL_AGGREGATION.items():
        try:
            df[v] = df[v].add(df.pop(k), fill_value=0)
        except KeyError as e:
            logger.debug(f"KeyError: {e}")

    order = [k for k in mp.DRAF_FUELS if k in df.keys()]
    return df[order]


def load_el_national_generation(
    year: int = 2019,
    country: str = "DE",
    freq: Optional[str] = "15min",
    cache: bool = True,
    split_queries: bool = True,
    ensure_std_techs: bool = True,
    ensure_std_index: bool = True,
    ensure_positive: bool = True,
    ensure_non_zero_sum: bool = True,
    fillna: bool = True,
    resample: bool = True,
) -> pd.DataFrame:

    assert year in range(2000, 2100), f"{year} is not a valid year"
    assert freq in [None, "15min", "30min", "60min"], f"{freq} is not a valid frequency"
    assert country in mp.EUROPE_COUNTRIES

    fp = paths.mode_dependent_cache_dir(year) / f"{year}_{country}_gen_entsoe.parquet"

    if cache and fp.exists():
        df = hp.read(fp)

    else:
        df = _query_generation(year, country, split_queries)

        if cache:
            hp.write(df, fp)

    data_freq = hp.estimate_freq(df)

    if ensure_std_index:
        idx = hp.make_datetimeindex(year, data_freq, tz=df.index.tz)
        df = df.reindex(idx)
        df = df.reset_index(drop=True)

    if ensure_std_techs:
        df = aggregate_to_standard_techs(df)

    if ensure_positive:
        # set negative values as missing values e.g. for HU and NL in 2019
        df[df < 0] = np.nan

    if ensure_non_zero_sum:
        df[df.sum(1) == 0.0] = np.nan

    if fillna:
        df = fill_special_missing_data_points_for_gen(df=df, country=country, year=year)
        df = df.fillna(method="ffill").fillna(method="bfill")

    if resample:
        df = hp.resample(df, year=year, start_freq=data_freq, target_freq=freq)

    return df


def _query_generation(year, country, split_queries) -> pd.DataFrame:
    client = _get_client()
    tz = get_timezone(country)
    if split_queries:
        df = _query_generation_monthly(year, country, client, tz)
    else:
        df = _query_generation_yearly(year, country, client, tz)
    return df


def _query_generation_yearly(year, country, client, tz) -> pd.DataFrame:
    try:
        start, end = _get_timestamps(year=year, tz=tz)
        df = client.query_generation(start=start, end=end, country_code=country)
    except (entsoe.exceptions.NoMatchingDataError, KeyError) as e:
        err_message = f"{e}: entsoe-client has no geneneration data found for {year, country}"
        logger.error(err_message)
        raise exceptions.NoDataError(err_message)
    return df


def fill_special_missing_data_points_for_gen(
    df: pd.DataFrame, country: str, year: int
) -> pd.DataFrame:
    if year == 2019 and country == "NL":
        logger.warning(f"No Data for 2019-01-01, NL. --> Data of next sunday 2019-01-06 is given.")
        st = hp.datetime_to_int("15min", 2019, 1, 6)
        df.iloc[:96] = df.iloc[st : st + 96].values

    if year == 2018 and country == "NL":
        logger.warning(f"No Data after 2018-12-31 23:00, NL. --> Data of previous hour is given.")
        df.iloc[35036:35040] = df.iloc[35032:35036].values

    if year == 2018 and country == "LT":
        logger.warning(f"1 outlier for {year}, {country}, other_conv/Other removed.")
        try:
            df["other_conv"] = hp.remove_outlier(df["other_conv"], zscore_threshold=10)
        except KeyError:
            df["Other"] = hp.remove_outlier(df["Other"], zscore_threshold=10)

    if year == 2018 and country == "LT":
        logger.warning(f"3 outlier for {year}, {country}, solar/Solar removed.")
        try:
            df["solar"] = hp.remove_outlier(df["solar"], zscore_threshold=10)
        except KeyError:
            df["Solar"] = hp.remove_outlier(df["Solar"], zscore_threshold=10)

    return df


def _query_generation_monthly(year, country, client, tz) -> pd.DataFrame:
    logger.warning(f"Querying generation data from entsoe for {country}: this may take minutes.")
    logger.info(f"Loading 12 month of electrcity generation individually:")

    d = {}
    for month in range(1, 13):

        start, end = _get_timestamps_for_month(year, tz, month)

        try:
            logger.warning(f"Loading month {month} / 12")
            d[month] = client.query_generation(start=start, end=end, country_code=country)
        except (entsoe.exceptions.NoMatchingDataError, KeyError) as e:
            raise exceptions.NoDataError(
                f"Entsoe-client has no geneneration data found for {year, country}: {e}"
            )
    return pd.concat(d.values(), sort=True)


def _get_timestamps_for_month(year, tz, month):
    start = pd.Timestamp(year=year, month=month, day=1, tz=tz)
    _year = year if month < 12 else year + 1
    _month = month + 1 if month < 12 else 1
    end = pd.Timestamp(year=_year, month=_month, day=1, tz=tz)
    return start, end


def prep_residual_load(
    year: int = 2019, freq: str = "15min", country: str = "DE", method: str = "all_conv",
) -> pd.Series:
    config = dict(year=year, freq=freq, country=country)

    if method == "all_conv":
        return get_conventional_generation(**config)

    elif method == "conv2":
        return get_conv2_generation(**config)

    # elif method == "complex":
    #     load = load_el_national_load(**config).fillna(0)
    #     imports = load_imports(**config).sum(1).fillna(0)
    #     exports = load_exports(**config).sum(1).fillna(0)
    #     res_ser = get_renewable_generation(**config).fillna(0)
    #     return load - imports + exports - res_ser

    else:
        raise ValueError(f"Method {method} not implemented.")


def get_subset_of(df, subset: Union[List, Set]) -> Tuple[pd.Series, Set]:
    wanted = set(subset)
    available = wanted.intersection(set(df.keys()))
    return df[available].sum(axis=1), available


def get_conventional_generation(year, freq, country) -> pd.Series:
    gen_TF = load_el_national_generation(year=year, freq=freq, country=country)
    conv_ser, available_conv = get_subset_of(gen_TF, mp.CONV)
    logger.info(f"Available CONV-fuels for {country}: {available_conv}")
    return conv_ser


def get_conv2_generation(year, freq, country) -> pd.Series:
    gen_TF = load_el_national_generation(year=year, freq=freq, country=country)
    conv_ser, available_conv = get_subset_of(gen_TF, mp.CONV2)
    logger.info(f"Available CONV-fuels for {country}: {available_conv}")
    return conv_ser


def get_renewable_generation(year, freq, country) -> pd.Series:
    gen_TF = load_el_national_generation(year=year, freq=freq, country=country)
    res_ser, available_vres = get_subset_of(gen_TF, mp.RES)
    logger.info(f"Available RES-fuels for {country}: {available_vres}")
    return res_ser


def load_el_national_specific_emissions() -> pd.DataFrame:
    """Read specific emissions data from [Tranberg.2019] in gCO2eq/kWh.

    [Tranberg.2019]: https://doi.org/10.1016/j.esr.2019.100367
    """
    fp = paths.DATA_DIR / f"tranberg/specific_emission_factors.csv"
    df = pd.read_csv(fp, header=0, index_col=0, skiprows=3).T
    return pd.DataFrame(
        {
            "wind_offshore": df["wind"],
            "wind_onshore": df["wind"],
            "nuclear": df["nuclear"],
            "other_RES": df["geothermal"],
            "biomass": df["biomass_cogeneration"],
            "pumped_hydro": df["hydropower_pumped_storage"],
            "hydro": df["hydropower_run-of-river"],
            "lignite": df["coal"],
            "other_conv": df["coal"],
            "coal": df["coal"],
            "gas": df["gas"],
            "oil": df["oil"],
            "solar": df["solar"],
        },
        index=df.index,
    ).T


def prep_shares(year: int = 2019, freq: str = "60min", country: str = "DE") -> pd.DataFrame:
    """Prepare shares of electricity production"""
    assert year in range(2000, 2100), f"{year} is not a valid year"
    assert freq in ["15min", "60min"], f"{freq} is not a valid freq"
    assert country in mp.EUROPE_COUNTRIES

    gen_TF = load_el_national_generation(year, country, freq="15min")
    gen_freq = hp.estimate_freq(gen_TF)

    gen_sums_T = gen_TF.sum(axis=1)
    shares_TF = gen_TF.copy()
    for f in shares_TF:
        shares_TF[f] = shares_TF[f] / gen_sums_T

    return hp.resample(shares_TF, year=year, start_freq=gen_freq, target_freq=freq)


def prep_dayahead_prices(
    year: int = 2019,
    freq: str = "60min",
    country: str = "DE",
    cache: bool = True,
    ensure_std_index: bool = True,
    fillna: bool = True,
    fill_outlier: bool = True,
    resample: bool = True,
) -> pd.Series:
    assert year in range(2000, 2100), f"{year} is not a valid year"
    assert country in entsoe_mp.BIDDING_ZONES

    fp = paths.CACHE_DIR / f"{year}_{freq}_{country}_dayahead_c_el_entsoe.parquet"

    bidding_zone = get_bidding_zone(country, year)

    if cache and fp.exists():
        ser = hp.read(fp)

    else:
        ser = _query_day_ahead_prices(year, bidding_zone)

        if cache:
            hp.write(ser, fp)

    if ensure_std_index:
        idx = hp.make_datetimeindex(year, hp.estimate_freq(ser), tz=ser.index.tz)
        ser = ser.reindex(idx).fillna(method="ffill").fillna(method="bfill")
        ser = ser.reset_index(drop=True)

    if fillna:
        ser = ser.fillna(ser.mean())

    if fill_outlier:
        ser = hp.fill_outlier_and_nan(ser, zscore_threshold=50)

    if resample:
        ser = hp.resample(ser, year=year, start_freq=hp.estimate_freq(ser), target_freq=freq)

    hp.warn_if_incorrect_index_length(ser, year, freq)
    return ser


def _query_day_ahead_prices(year, bidding_zone) -> pd.Series:
    client = _get_client()
    tz = get_timezone(bidding_zone)
    try:
        start, end = _get_timestamps(year=year, tz=tz)
        ser = client.query_day_ahead_prices(start=start, end=end, country_code=bidding_zone)
    except entsoe.exceptions.NoMatchingDataError as e:
        raise exceptions.NoDataError(
            f"Entsoe-client has no price-data found for {year, bidding_zone}: {e}"
        )
    return ser


def get_bidding_zone(country: str, year: int) -> str:
    if year == 2018 and country == "IE":
        logger.warning(
            "Ireland coupled with GB on 2018-10-01 and changed from 30min to 60min Day-Ahead-Markt"
            " prices. See https://www.nordpoolgroup.com/message-center-container/newsroom/feature/2018/10/nord-pool-welcomes-power-coupling-with-ireland/"
        )
    if year <= 2018 and country == "AT":
        country = "DE"
        logger.info("AT and DE have the same prices.")

    if country in ["DE", "LU"]:
        if year > 2018:
            country = "DE-LU"
        else:
            country = "DE-AT-LU"

    country_to_bidding = {
        "NO": "NO-1",
        "DK": "DK-2",
        "IT": "IT-NORD",
        "SE": "SE-1",
        "IE": "IE-SEM",
    }
    return country_to_bidding.get(country, country)


def get_timezone(country: str) -> str:
    return entsoe_mp.TIMEZONE_MAPPINGS.get(country, "Europe/Brussels")


#############################
# CURRENTLY UNUSED FUNCTIONS
#############################

# def load_imports(
#     year: int = 2019,
#     country: str = "DE",
#     freq: Optional[str] = "15min",
#     cache: bool = True,
#     ensure_std_index: bool = True,
# ) -> pd.DataFrame:

#     # see visualization of the ENTSO-E data under: https://www.energy-charts.de/exchange.htm

#     assert year in range(2000, 2100), f"{year} is not a valid year"
#     assert freq in [None, "15min", "30min", "60min"], f"{freq} is not a valid frequency"

#     aggregations = {
#         "DK": ["DK-1", "DK-2"],
#         "NO": ["NO-1", "NO-2", "NO-3", "NO-4", "NO-5"],
#         "SE": ["SE-1", "SE-2", "SE-3", "SE-4"],
#     }

#     if country in aggregations:
#         config = dict(year=year, freq=freq, ensure_std_index=ensure_std_index)
#         return sum([load_imports(country=region, **config) for region in aggregations[country]])

#     fp = paths.CACHE_DIR / f"{year}_{country}_import_entsoe.h5"

#     bidding_zone = get_bidding_zone(country=country, year=year)

#     if cache and fp.exists():
#         df = pd.read_hdf(fp)

#     else:
#         client = _get_client()
#         tz = get_timezone(bidding_zone)
#         start, end = _get_timestamps(year=year, tz=tz)
#         df = client.query_import(start=start, end=end, country_code=bidding_zone)
#         if cache:
#             hp.write(df, fp)

#     if ensure_std_index:
#         idx = hp.make_datetimeindex(year, hp.estimate_freq(df), tz=df.index.tz)
#         df = df.reindex(idx).fillna(method="ffill").fillna(method="bfill")

#     df = hp.resample(df, year=year, start_freq=hp.estimate_freq(df), target_freq=freq)

#     return df


# def load_crossborder_flows(
#     country_from: str,
#     country_to: str,
#     year: int = 2019,
#     #    freq: Optional[str] = "15min",
#     cache: bool = True,
#     #    ensure_std_index: bool = True
# ) -> pd.DataFrame:

#     assert year in range(2000, 2100), f"{year} is not a valid year"
#     # assert freq in [None, "15min", "30min", "60min"], f"{freq} is not a valid frequency"

#     fp = paths.CACHE_DIR / f"{year}_crossborder_{country_from}--{country_to}_entsoe.h5"

#     if cache and fp.exists():
#         ser = pd.read_hdf(fp)

#     else:
#         client = _get_client()
#         tz = get_timezone(get_bidding_zone(country=country_from, year=year))
#         start, end = _get_timestamps(year=year, tz=tz)

#         ser = client.query_crossborder_flows(country_from, country_to, start=start, end=end)
#         if cache:
#             hp.write(ser, fp)

#     # if ensure_std_index:
#     #     idx = hp.make_datetimeindex(year, hp.estimate_freq(ser), tz=ser.index.tz)
#     #     ser = ser.reindex(idx).fillna(method="ffill").fillna(method="bfill")

#     # ser = hp.resample(ser, year=year, start_freq=hp.estimate_freq(ser), target_freq=freq)

#     return ser


# def load_exports(
#     year: int = 2019,
#     country: str = "DE",
#     freq: Optional[str] = "15min",
#     cache: bool = True,
#     ensure_std_index: bool = True,
# ) -> pd.DataFrame:

#     assert year in range(2000, 2100), f"{year} is not a valid year"
#     assert freq in [None, "15min", "30min", "60min"], f"{freq} is not a valid frequency"

#     fp = paths.CACHE_DIR / f"{year}_{country}_export_entsoe.h5"

#     bzone_from = get_bidding_zone(country=country, year=year)

#     if cache and fp.exists():
#         df = pd.read_hdf(fp)

#     else:
#         client = _get_client()
#         export_dic = {}
#         for bzone_to in entsoe_mp.NEIGHBOURS[bzone_from]:
#             tz = get_timezone(country=bzone_to)
#             start, end = _get_timestamps(year=year, tz=tz)
#             try:
#                 export_dic[bzone_to] = client.query_crossborder_flows(
#                     country_code_from=bzone_from,
#                     country_code_to=bzone_to,
#                     start=start,
#                     end=end,
#                     lookup_bzones=True,
#                 )
#             except entsoe.exceptions.NoMatchingDataError:
#                 logger.warning(f"NoMatchingDataError: No data for export to {bzone_to}")

#         df = pd.concat(export_dic, axis=1)
#         if cache:
#             hp.write(df, fp)

#     if ensure_std_index:
#         idx = hp.make_datetimeindex(year, hp.estimate_freq(df), tz=df.index.tz)
#         df = df.reindex(idx).fillna(method="ffill").fillna(method="bfill")

#     df = hp.resample(df, year=year, start_freq=hp.estimate_freq(df), target_freq=freq)

#     return df


# def load_el_national_load(
#     year: int = 2019,
#     freq: str = "15min",
#     country: str = "DE",
#     cache: bool = True,
#     ensure_std_index: bool = True,
#     resample: bool = True,
# ) -> pd.Series:
#     assert year in range(2000, 2100), f"{year} is not a valid year"
#     assert freq in [None, "15min", "30min", "60min"], f"{freq} is not a valid freq"
#     assert country in mp.EUROPE_COUNTRIES or country in entsoe_mp.BIDDING_ZONES

#     fp = paths.CACHE_DIR / f"{year}_{country}_load_entsoe.h5"

#     if cache and fp.exists():
#         ser = pd.read_hdf(fp)

#     else:
#         if country == "DE" and year == 2018:
#             a = load_el_national_load(year=year, freq=freq, country="DE-AT-LU")
#             b = load_el_national_load(year=year, freq=freq, country="DE-LU")
#             idx = hp.make_datetimeindex(year=year, freq=freq, tz=a.index.tz)
#             ser = pd.concat([a, b], axis=1).mean(1)

#         else:
#             client = _get_client()
#             tz = get_timezone(country)
#             try:
#                 start, end = _get_timestamps(year=year, tz=tz)
#                 ser = client.query_load(start=start, end=end, country_code=country)
#             except entsoe.exceptions.NoMatchingDataError as e:
#                 raise exceptions.NoDataError(
#                     f"Entsoe-client has no load-data found for {year, country}: {e}"
#                 )

#         if cache:
#             hp.write(ser, fp)

#     if ensure_std_index:
#         idx = hp.make_datetimeindex(year, hp.estimate_freq(ser), tz=ser.index.tz)
#         ser = ser.reindex(idx).fillna(method="ffill").fillna(method="bfill").reset_index(drop=True)

#     if resample:
#         ser = hp.resample(ser, year=year, start_freq=hp.estimate_freq(ser), target_freq=freq)
#     return ser



import logging
import os
from datetime import datetime
from functools import lru_cache
from pathlib import Path
from typing import Any, Callable, Dict, Iterable, List, Optional, Tuple, Union

import numpy as np
import pandas as pd
from numpy.core.fromnumeric import squeeze

from elmada import paths

logger = logging.getLogger(__name__)
logger.setLevel(level=logging.WARN)

DEFAULT_HEADER = "DEFAULT_HEADER"

APIS = {
    "entsoe": (
        "ENTSO-E API key",
        "https://transparency.entsoe.eu/content/static_content/Static%20content/web%20api/Guide.html",
        "ENTSOE_API_KEY",
    ),
    "morph": ("Morph API key", "https://morph.io/documentation/api", "MORPH_API_KEY"),
    "quandl": (
        "Quandl API key",
        "https://docs.quandl.com/docs#section-authentication",
        "QUANDL_API_KEY",
    ),
}


def make_symlink_to_cache():
    """Creates a symbolic link to the cache directory for easy access.

    Note: This function requires admin privileges.
    """
    link_dir = paths.BASE_DIR / "cache"
    cache_dir = paths.CACHE_DIR
    link_dir.symlink_to(target=cache_dir, target_is_directory=True)
    print(f"Symbolic link created: {link_dir} --> {cache_dir}")


def delete_cache(filter_str: str = "*") -> None:
    """Deletes parts or all of the cache directory.

    Unless `filter_str` is '*' only files are selected that contain the `filter_string` somewhere
    in the filename.
    """

    s = "*" if filter_str == "*" else f"*{filter_str}*"

    files = list(paths.CACHE_DIR.glob(f"{s}"))
    lenf = len(files)

    if lenf == 0:
        print(f"No file found containing '{filter_str}'.")

    else:
        print(f"{lenf} files containing '{filter_str}':")

        for f in files:
            size = sizeof_fmt(f.stat().st_size)
            print(f"\t{size:>5}{f.name}")

        if confirm_deletion(lenf):
            for f in files:
                f.unlink()
            print(f"{lenf} files deleted")

        else:
            print("No files deleted")


def confirm_deletion(nfiles: int) -> bool:
    return input(f"Do you really want to delete these {nfiles} files? (y/n)") == "y"


def print_error(series1: pd.Series, series2: pd.Series) -> None:
    err = abs(series1 - series2)
    err_rel = err / series1
    print(f"relative error = {err_rel.mean():.2%}, absolute error= {err.mean():.2f} t CO2/MWh.")


def add_row_for_steps(ser: pd.Series) -> pd.Series:
    """Adds an initial row in the Series for plotting steps. This way the first step does not
    disappear when choosing option `pre`.
    """
    return ser.append(pd.Series({-1: 0})).sort_index()


def sizeof_fmt(num: float, suffix: str = "B") -> str:
    """Returns the short version of the storage size of e.g. digital information.

    Example:
        sizeof_fmt(5000) -> 5.0 KB

    Idea from https://stackoverflow.com/questions/1094841
    """
    for unit in ["", "K", "M", "G", "T", "P", "E", "Z"]:
        if abs(num) < 1000.0:
            return f"{num:.1f} {unit}{suffix}"
        num /= 1000.0
    return f"{num:.1f} Y{suffix}"


def freq_to_hours(freq: str) -> float:
    """E.g.:
    '60min' -> 1.0
    '30min' -> 0.5
    """
    assert freq in ["60min", "30min", "15min"]
    return float(freq[:2]) / 60


def int_to_datetime(freq: str, year: int, pos: int) -> "datetime":
    dtindex = make_datetimeindex(year=year, freq=freq)
    return dtindex[pos]


def datetime_to_int(freq: str, year: int, month: int, day: int) -> int:
    """Returns the index location in a whole-year date-time-index for the given date."""
    dtindex = make_datetimeindex(year=year, freq=freq)
    i = datetime(year, month, day)
    return dtindex.get_loc(i)


def is_correct_length(df: Union[pd.DataFrame, pd.Series], year: int, freq: str):
    len_dt = len(make_datetimeindex(year=year, freq=freq))
    len_inp = len(df)
    return len_dt == len_inp


def write(data: Union[pd.Series, pd.DataFrame], fp: Union[Path, str]) -> None:
    """Standardized way of writing arrays in draf.

    Args:
        data: Must be either pandas.Series or pandas.DataFrame.
        fp: Filepath of a .parquet or .csv file.
    """

    fp = Path(fp)

    if fp.suffix == ".parquet":
        if isinstance(data, pd.Series):
            # NOTE: Parquet can only write dataframes, not series
            if data.name is None:
                data.name = DEFAULT_HEADER
        pd.DataFrame(data).to_parquet(fp, index=True)
    # elif fp.suffix == ".h5":
    #     data.to_hdf(fp, key="default", mode="w")
    elif fp.suffix == ".csv":
        data.to_csv(fp, index=True)
    else:
        raise ValueError(f"Suffix {fp.suffix} not supported")


def read(fp: Union[Path, str], squeeze: bool = True) -> Union[pd.Series, pd.DataFrame]:
    """Standardized way of reading arrays in draf.

    Args:
        fp: Filepath of a .parquet, or .csv file.
        squeeze: If DataFrames with one columns should be transformed into a series
    """

    fp = Path(fp)

    if fp.suffix == ".parquet":
        data = pd.read_parquet(fp)
    # elif fp.suffix == ".h5":
    #     data = pd.read_hdf(fp)
    elif fp.suffix == ".csv":
        data = pd.read_csv(fp, index_col=0)
    else:
        raise ValueError(f"Suffix {fp.suffix} not supported")

    if squeeze:
        data = data.squeeze()

        try:
            if data.name == DEFAULT_HEADER:
                data.name = None
        except (AttributeError, ValueError):
            # AttributeError raised if data is a DataFrame, except when also column 'name' exist,
            # then a ValueError is raised
            pass

    return data


def warn_if_incorrect_index_length(
    df: Union[pd.DataFrame, pd.Series], year: int, freq: str
) -> None:
    len_dt = len(make_datetimeindex(year=year, freq=freq))
    len_inp = len(df)
    match = len_dt == len_inp
    if not match:
        logger.warning(f"Length mismatch: An object has {len_inp} instead of {len_dt}.")


@lru_cache(maxsize=16)
def make_datetimeindex(year: int = 2019, freq: str = "60min", tz: str = None) -> pd.DatetimeIndex:
    """Returns a whole-year date-time-index in desired resolution."""
    date_start = f"01.01.{year}  00:00:00"
    date_end = f"31.12.{year}  23:59:00"
    return pd.date_range(date_start, date_end, freq=freq, tz=tz)


def resample(
    df: Union[pd.Series, pd.DataFrame],
    year: int,
    start_freq: str,
    target_freq: Optional[str],
    aggfunc: str = "mean",
) -> Union[pd.Series, pd.DataFrame]:
    """Resamples data from a start frequency to a target frequency."""
    if (target_freq is None) or (start_freq == target_freq):
        return df
    else:
        if int_from_freq(start_freq) > int_from_freq(target_freq):
            func = upsample
        else:
            func = downsample

        return func(
            df=df, year=year, start_freq=start_freq, target_freq=target_freq, aggfunc=aggfunc,
        )


def downsample(
    df: Union[pd.Series, pd.DataFrame],
    year: int,
    start_freq: str = "15min",
    target_freq: str = "60min",
    aggfunc: str = "mean",
) -> Union[pd.Series, pd.DataFrame]:
    """Downsampling for cases where start frequency is higher then target frequency.

    Args:
        df: Series or Dataframe
        year: Year
        start_freq: Time resolution of given data.
        target_freq: Time resolution of returned data.
        aggfunc:  in {"mean", "sum"}
    """

    df = df.copy()
    df.index = make_datetimeindex(year, freq=start_freq)

    resampler = df.resample(target_freq)

    if aggfunc == "sum":
        df = resampler.sum()
    elif aggfunc == "mean":
        df = resampler.mean()
    else:
        raise ValueError(f"Aggfunc '{aggfunc}' not valid")

    df = df.reset_index(drop=True)

    if isinstance(df, pd.Series) and isinstance(df.name, str):
        df.name = df.name.replace(start_freq, target_freq)

    warn_if_incorrect_index_length(df, year, target_freq)
    return df


def upsample(
    df: Union[pd.Series, pd.DataFrame],
    year: int,
    start_freq: str = "60min",
    target_freq: str = "15min",
    aggfunc: str = "mean",
) -> Union[pd.Series, pd.DataFrame]:
    """Upsampling for cases where start frequency is lower then target frequency.

    Args:
        df: Series or Dataframe
        year: Year
        start_freq: Time resolution of given data.
        target_freq: Time resolution of returned data.
        aggfunc: Either 'mean' or 'sum', e.g. use `mean` for power and `sum` for aggregated energy.
    """
    df = df.copy()
    df.index = make_datetimeindex(year, freq=start_freq)
    df = df.resample(target_freq).pad()
    convert_factor = int_from_freq(start_freq) / int_from_freq(target_freq)
    if aggfunc == "sum":
        df /= convert_factor

    df = df.reset_index(drop=True)
    df = _append_rows(df, convert_factor=convert_factor)

    if isinstance(df, pd.Series) and isinstance(df.name, str):
        df.name = df.name.replace(start_freq, target_freq)

    warn_if_incorrect_index_length(df, year, target_freq)
    return df


def _append_rows(
    df: Union[pd.Series, pd.DataFrame], convert_factor: float
) -> Union[pd.Series, pd.DataFrame]:
    length = len(df)
    if isinstance(df, pd.DataFrame):
        addon_data = [df.iloc[length - 1]]
    elif isinstance(df, pd.Series):
        addon_data = df.iloc[length - 1]
    else:
        raise RuntimeError(f"Unsupported type {type(df)}")

    number_of_lines = int(convert_factor - 1)
    new_index = [length + i for i in range(number_of_lines)]

    class_of_df = type(df)
    line = class_of_df(data=addon_data, index=new_index)  # new Series OR Dataframe.
    df = df.append(line).sort_index().reset_index(drop=True)
    return df


def z_score(df: Union[pd.Series, pd.DataFrame]) -> Union[pd.Series, pd.DataFrame]:
    """Returns the standard score for all data points.
    Info: The z-score tells how many standard deviations below or above the population mean a raw
     score is.
    """
    return (df - df.mean()) / df.std(ddof=0)


def remove_outlier(
    df: Union[pd.Series, pd.DataFrame], zscore_threshold: float = 2.698
) -> Union[pd.Series, pd.DataFrame]:
    """Detects and removes outliers of Dataframes or Series.

    Note: The default z-score of 2.698 complies to the boxplot standard (1,5*IQR-rule).
     (see https://towardsdatascience.com/5e2df7bcbd51)
    """
    df = df.copy()

    # see
    # https://stackoverflow.com/questions/23451244/how-to-zscore-normalize-pandas-column-with-nans

    if isinstance(df, pd.DataFrame):
        for k in df:
            df[k] = remove_outlier(df[k], zscore_threshold)

    if isinstance(df, pd.Series):
        outliers = np.abs(z_score(df)) > zscore_threshold
        n_outliers = outliers.sum()
        df[outliers] = np.nan
        outlier_share = n_outliers / len(df)
        outlier_treshold = 0.1
        if outlier_share > outlier_treshold:
            logger.warning(
                f"{outlier_share:.2%} (more than {outlier_treshold:.0%}) of the data were outliers"
            )
        logger.info(f"{n_outliers} datapoints where removed in {df.name}")

    return df


def fill_outlier_and_nan(
    df: Union[pd.Series, pd.DataFrame], zscore_threshold: int = 3, method: str = "linear",
) -> Union[pd.Series, pd.DataFrame]:

    df = remove_outlier(df, zscore_threshold)

    if method == "linear":
        df = df.interpolate(method="linear")
    elif method == "fill":
        df = df.fillna(method="bfill").fillna(method="ffill")

    return df


def estimate_freq(data: Union[List, pd.Series, pd.DataFrame]) -> str:
    """Estimate the correct frequency by the length of the yearly data-set."""
    data_len = len(data)
    minutes_per_year = 8760 * 60
    tolerance = 0.2

    for step_length_in_minutes in [60, 30, 15]:
        error = abs(data_len * step_length_in_minutes / minutes_per_year - 1)
        if error < tolerance:
            return f"{step_length_in_minutes}min"

    raise RuntimeError(f"Data contains {data_len} datapoints. It cannot be matched to a frequency.")


def int_from_freq(freq: str) -> int:
    """E.g. '15min' -> 15"""
    return int(freq[:2])


def set_api_keys(**kwargs) -> None:
    for which, api_key in kwargs.items():
        set_api_key(which=which, api_key=api_key)


def set_api_key(which: str, api_key: str) -> None:
    fp = paths.KEYS_DIR / f"{which}.txt"
    assert which in APIS, f"`which` must be one of {list(APIS.keys())}."
    if fp.exists():
        respond = input(
            f"The file {fp} already exists. Do you want to overwrite it? Type (y)es or (n)o."
        )
        if respond not in ["yes", "y"]:
            print("Aborted.")
            return None

    fp.write_text(api_key)
    print(f"Api key written to {fp}.")


def get_api_key(which: str = "entsoe"):
    assert which in APIS, f"`which` must be one of {list(APIS.keys())}."

    var_name = APIS[which][2]

    try:
        return os.environ[var_name]

    except KeyError:

        try:
            fp = paths.KEYS_DIR / f"{which}.txt"
            return fp.read_text().strip()

        except FileNotFoundError as e:
            raise Exception(
                f"`{which}.txt` file not found and {var_name} not set."
                f"Please get a valid {APIS[which][0]} "
                f"(see {APIS[which][1]}) and place it in `elmada/api_keys/{which}.txt`\n"
            ) from e



import logging
from typing import Any, Callable, Dict, Iterable, List, Optional, Tuple, Union

import pandas as pd

import elmada
from elmada import helper as hp
from elmada import paths

logger = logging.getLogger(__name__)
logger.setLevel(level=logging.WARN)


def get_emissions(
    year: int,
    freq: str = "60min",
    country: str = "DE",
    method: str = "XEF_PP",
    cache: bool = True,
    use_datetime: bool = False,
    **mo_kwargs,
) -> Union[pd.Series, pd.DataFrame]:
    """Returns dynamic carbon emisson factors in gCO2eq/kWh_el and optional data.

    Args:
        year: Year
        freq: Frequency, e.g. '60min' or '15min'
        country: alpha-2 country code, e.g. 'DE'
        method: A method string that consists of two parts joined by an underscore. The first
            part is the type of emission factor. Use 'XEF' for grid mix emission factors and
            'MEF' for marginal emission factors. The second part determines the calculation method
            or data sources. 'EP' stands for ENTSO-E, 'PP' for power plant method, 'PWL' for piecewise
            linear method, 'PWLv' for the piecewise linear method in validation mode.
            For PP, PWL, and PWLv, the first can be omitted ('_PP', '_PWL', '_PWLv') to return
            a DataFrame including the columns 'residual_load', 'total_load', 'marginal_fuel',
            'efficiency', 'marginal_cost', 'MEFs', 'XEFs'.
            The following combinations are possible:

            | Method   | Description                              |
            | -------- | ---------------------------------------- |
            | XEF_EP   | Series: XEFs using ENTSO-E data          |
            | XEF_PP   | Series: XEFs using PP method             |
            | XEF_PWL  | Series: XEFs using PWL method            |
            | XEF_PWLv | Series: XEFs using PWLv method           |
            | MEF_PP   | Series: MEFs from PP method              |
            | MEF_PWL  | Series: MEFs using PWL method            |
            | MEF_PWLv | Series: MEFs using PWLv method           |
            | _PP      | Dataframe: extended data for PP method   |
            | _PWL     | Dataframe: extended data for PWL method  |
            | _PWLv    | Dataframe: extended data for PWLv method |

        cache: If cache is used.
        use_datetime: If True, the index is a timezone agnostic datetime. If False, the index is
            0, 1, 2, etc.
        **mo_kwargs: Keyword arguments for merit order creation such as 'overwrite_carbon_tax',
            'efficiency_per_plant', 'emission_data_source'.
    """
    first_method_part, last_method_part = method.split("_")

    fp = paths.CACHE_DIR / f"{year}_{country}_{freq}_CEFs_{last_method_part}.parquet"

    if cache and fp.exists() and not mo_kwargs:
        df = hp.read(fp)
    else:
        df = _make_emissions(
            year=year, freq=freq, country=country, method=last_method_part, **mo_kwargs
        )
        if cache:
            hp.write(df, fp)

    if use_datetime:
        df.index = hp.make_datetimeindex(year=year, freq=freq)

    if first_method_part in ("XEF", "MEF"):
        return df[first_method_part + "s"]
    else:
        return df


def _make_emissions(year, freq, country, method, **mo_kwargs) -> pd.DataFrame:
    config = dict(year=year, freq=freq, country=country)

    if method == "EP":
        return elmada.from_entsoe.prep_XEFs(**config)
    # elif method == "SM":
    #     return elmada.from_smard.prep_XEFs(**config)
    elif method == "PP":
        return elmada.from_opsd.prep_CEFs(**config, **mo_kwargs)
    elif method in ["PWL", "PWLv"]:
        is_vmode = bool(method == "PWLv")
        return elmada.eu_pwl.prep_CEFs(validation_mode=is_vmode, **config, **mo_kwargs)
    else:
        raise ValueError(f"Method {method} not implemented.")


def get_prices(
    year: int,
    freq: str = "60min",
    country: str = "DE",
    method: str = "hist_EP",
    cache: bool = True,
    **mo_kwargs,
) -> pd.Series:
    """Returns the day-ahead spot-market electricity prices in €/MWh.

    Args:
        year: Year
        freq: Frequency, e.g. '60min' or '15min'
        country: alpha-2 country code, e.g. 'DE'
        method: 'PP' stands for power plant method, 'PWL' stands for piecewise
            linear method, 'PWLv' stands for the piecewise linear method in validation mode.

            | Method  | Description                                         |
            | ------- | --------------------------------------------------- |
            | PP      | Using PP method                                     |
            | PWL     | Using PWL method                                    |
            | PWLv    | Using PWLv method                                   |
            | hist_EP | Using historic ENTSO-E data                         |
            | hist_SM | Using historic Smard data only for DE, (2015, 2018  |

        cache: If data is cached

    Known data issues:
        - (2015, DE, entsoe)-prices: data missing until Jan 6th
        - (2018, DE, entsoe)-prices: data missing from Sep 30th
    """

    # >> Workaround due to missing data for DE
    if year in [2015, 2018] and country == "DE" and method == "hist_EP":
        method = "hist_SM"
        logger.warning(
            f"The requested entsoe-data ({year}, {country}) is not complete. "
            f"Smard-data is given to you instead."
        )
    # <<

    config = dict(year=year, freq=freq, country=country, cache=cache)

    if method == "hist_EP":
        return elmada.from_entsoe.prep_dayahead_prices(**config)
    elif method == "hist_SM":
        return elmada.from_smard.prep_dayahead_prices(**config)
    elif method in ["PP", "PWL", "PWLv"]:
        df = get_emissions(method=f"_{method}", **config, **mo_kwargs)
        return df["marginal_cost"]
    else:
        raise ValueError(f"Method '{method}' not implemented.")


def get_merit_order(
    year: int, country: str = "DE", method: str = "PP", **mo_kwargs
) -> pd.DataFrame:
    """Returns the merit order as DataFrame.

    Args:
        year: Year
        country: alpha-2 country code, e.g. 'DE'
        method: One of 'PP' (power plant method), 'PWL' (piecewise
            linear method), 'PWLv' (piecewise linear method in validation mode).
        **mo_kwargs: keyword arguments for merit order function depending on `method`.
    """
    if method == "PP":
        assert country == "DE", f"PP-method only works for Germany and not for {country}"
        return elmada.from_opsd.merit_order(year=year, **mo_kwargs)
    elif method == "PWL":
        return elmada.eu_pwl.merit_order(
            year=year, country=country, validation_mode=False, **mo_kwargs
        )
    elif method == "PWLv":
        return elmada.eu_pwl.merit_order(
            year=year, country=country, validation_mode=True, **mo_kwargs
        )
    else:
        raise ValueError("`method` needs to be one of ['PP', 'PWL', 'PWLv'].")


def get_residual_load(year: int, freq: str = "60min", country: str = "DE", **kwargs) -> pd.Series:
    return elmada.from_entsoe.prep_residual_load(year=year, freq=freq, country=country, **kwargs)


def get_el_national_generation(year: int, freq: str = "60min", country: str = "DE") -> pd.DataFrame:
    return elmada.from_entsoe.load_el_national_generation(year=year, freq=freq, country=country)



"""Reads power plant list from Open Power System Data (OPSD)."""

import logging
from pathlib import Path
from typing import Any, Callable, Dict, Iterable, List, Optional, Tuple, Union

import numpy as np
import pandas as pd
import requests
from scipy import stats

import elmada
from elmada import from_entsoe, from_other
from elmada import mappings as mp
from elmada import paths

logger = logging.getLogger(__name__)
logger.setLevel(logging.WARN)


def prep_prices(year=2019, freq="60min", country="DE", **mo_kwargs) -> pd.Series:
    """Convenience function to get the marginal costs for each timesteps."""
    return prep_CEFs(year=year, freq=freq, country=country, **mo_kwargs)["marginal_cost"]


def prep_CEFs(year=2019, freq="60min", country="DE", mo_P=None, **mo_kwargs) -> pd.DataFrame:
    """Prepares German CEFs from the power plant list from OPSD"""
    assert country == "DE", "this function only works for Germany"
    if mo_P is None:
        mo_P = merit_order(year=year, **mo_kwargs)
    return get_CEFs_from_merit_order(mo_P=mo_P, year=year, freq=freq, country=country)


def get_CEFs_from_merit_order(
    mo_P: pd.DataFrame, year: int, freq: str, country: str, resi_T: Optional[pd.Series] = None,
) -> pd.DataFrame:
    if resi_T is None:
        resi_T = from_entsoe.prep_residual_load(year=year, freq=freq, country=country)
    _warn_if_not_enough_capa(mo_P, resi_T)
    total_load_T = from_entsoe.load_el_national_generation(
        year=year, freq=freq, country=country
    ).sum(axis=1)
    cols = [
        "cumsum_capa",
        "marginal_emissions",
        "capa",
        "fuel_draf",
        "used_eff",
        "marginal_cost",
    ]
    len_mo = len(mo_P)
    mo_P_arr = mo_P[cols].values
    transm_eff = from_other.get_transmission_efficiency(country=country)

    def get_data(resi_value: float, what: str = "marginal_emissions"):
        output_col_index = cols.index(what)
        for row in range(len_mo):
            if mo_P_arr[row, 0] > resi_value:
                return mo_P_arr[row, output_col_index]
        return mo_P_arr[-1, output_col_index]

    def get_emissions_for_xef(resi_value: float) -> float:
        # If the residual load is not positive there are no emissions:
        if resi_value <= 0.0:
            return 0.0

        emissions = 0.0
        for row in range(len_mo):
            emissions += mo_P_arr[row, 1] * mo_P_arr[row, 2]

            # If the cumsum_capa is above resi_value subtract the emissions of the additional
            # capacity and return the quotient of emissions and resi_value:
            if mo_P_arr[row, 0] > resi_value:
                emissions -= mo_P_arr[row, 1] * (mo_P_arr[row, 0] - resi_value)
                return emissions

        # If the following code is executed there is not enough generation capacity:
        return emissions

    df = pd.DataFrame(
        {
            "residual_load": resi_T,
            "total_load": total_load_T,
            "marginal_fuel": resi_T.apply(get_data, what="fuel_draf"),
            "efficiency": resi_T.apply(get_data, what="used_eff"),
            "marginal_cost": resi_T.apply(get_data, what="marginal_cost"),
            "MEFs": resi_T.apply(get_data, what="marginal_emissions") / transm_eff,
            "XEFs": (resi_T.apply(get_emissions_for_xef)) / total_load_T / transm_eff,
        }
    )

    df["XEFs"] *= 1000  # convert from t/MWh to kg/MWh or g/kWh
    df["MEFs"] *= 1000  # convert from t/MWh to kg/MWh or g/kWh
    return df


def merit_order(
    year=2019,
    efficiency_per_plant: bool = True,
    emission_data_source: str = "quaschning",
    overwrite_carbon_tax: Optional[float] = None,
    **preprocess_kwargs,
) -> pd.DataFrame:
    """Prepares the merit order from the German power plant list."""

    df = get_current_active_power_plants(year)
    df = _rename_to_draf_fuels(df)
    df = _preprocess_efficiencies(
        df, efficiency_per_plant=efficiency_per_plant, **preprocess_kwargs
    )
    df = _add_marginal_emissions_for_gen(df, emission_data_source)

    carbon_price = (
        from_other.get_ETS_price(year) if overwrite_carbon_tax is None else overwrite_carbon_tax
    )
    df["x_k"] = df["fuel_draf"].map(from_other.get_fuel_prices(year=year, country="DE"))
    df["fuel_cost"] = df["x_k"] / df["used_eff"]
    df["GHG_cost"] = df["marginal_emissions_for_gen"] * carbon_price
    df["marginal_emissions"] = df["marginal_emissions_for_gen"]
    df["marginal_cost"] = df["fuel_cost"] + df["GHG_cost"]

    # to be consistent with the PWL merit_order:
    df = df.rename(columns={"capacity_net_bnetza": "capa"})

    df = df.sort_values("marginal_cost").reset_index(drop=True)
    df["cumsum_capa"] = df["capa"].cumsum()
    return df


def get_summary(year=2019) -> pd.DataFrame:
    ca = get_current_active_power_plants(year)
    grouper = ca.groupby(["fuel", "technology"])
    d = dict(
        counts=grouper.count().id,
        efficiency=grouper.mean()["efficiency_estimate"],
        sum_capa=grouper.sum()["capacity_net_bnetza"],
    )
    return pd.concat(d.values(), axis=1, keys=d.keys())


def get_summary_of_opsd_raw() -> pd.DataFrame:
    df = read_opsd_powerplant_list(which="DE")
    df = df.rename(columns={"capacity_net_bnetza": "capa"})
    df = df.groupby("energy_source_level_2").agg({"id": "size", "capa": "sum"})
    df = df.sort_values("capa", ascending=False)
    df["capa_rel"] = df["capa"] / df["capa"].sum()
    return df


def get_current_active_power_plants(year=2019) -> pd.DataFrame:
    """Returns a Dataframe with all active power plants for a specific year."""
    df = read_opsd_powerplant_list(which="DE")

    used_fuels = mp.OPSD_TO_DRAF.keys()
    other_fuels = set(df["fuel"]) - set(used_fuels)
    is_german = df["country_code"] == "DE"
    after_commissioned = (df["commissioned"].notnull() & (df["commissioned"] < year)) | df[
        "commissioned"
    ].isna()
    before_shutdown = (df["shutdown"].notnull() & (df["shutdown"] > year)) | df["shutdown"].isna()
    currently_active = after_commissioned & before_shutdown
    in_used_fuels = df["fuel"].isin(used_fuels)
    cond = is_german & currently_active & in_used_fuels
    ca = df[cond]

    logger.info(
        f"{(cond.sum() / len(df)):.2%} of data rows were used "
        f"({(~in_used_fuels).sum() / len(df):.2%} are not in used fuels). "
        f"Other (discarded) fuels are {other_fuels}."
    )

    # interesting_cols = ["id", "name_bnetza", "status", "fuel", "technology", "type", "efficiency_estimate", "capacity_net_bnetza"]
    return ca


def read_opsd_powerplant_list(which: str = "DE") -> pd.DataFrame:
    assert which in ("DE", "EU"), f"`{which}` is no valid value for `which`."

    fp = paths.mode_dependent_cache_dir() / f"OPSD_conventional_power_plants_{which}.csv"

    if not fp.exists():
        download_powerplant_list(which=which, fp=fp)
    df = pd.read_csv(fp)

    if elmada.get_mode() == "live":
        df = df.rename(columns={"energy_source": "fuel", "country": "country_code"})

    return df


def download_powerplant_list(which: str, fp: Path) -> None:
    url = (
        f"https://data.open-power-system-data.org/conventional_power_plants/latest/"
        f"conventional_power_plants_{which}.csv"
    )
    response = requests.get(url, allow_redirects=True)
    fp.write_bytes(response.content)
    logger.info(f"{fp.name} downloaded from OPSD.")


def _rename_to_draf_fuels(
    df: pd.DataFrame, minimum_efficiency_for_gas_cc: float = 0.5
) -> pd.DataFrame:
    def my_rename(row):
        if (
            (row["fuel"] == "Natural gas")
            and (row["technology"] == "Combined cycle")
            and (row["efficiency_estimate"] >= minimum_efficiency_for_gas_cc)
        ):
            return "gas_cc"
        else:
            return mp.OPSD_TO_DRAF[row["fuel"]]

    df["fuel_draf"] = df.apply(my_rename, axis=1)
    return df


def _preprocess_efficiencies(
    df: pd.DataFrame,
    efficiency_per_plant: bool = True,
    fill_missing_efficiencies: bool = True,
    fill_zscore_outlier: bool = True,
    zscore_threshold: float = 3,
    ensure_minimum_efficiency: bool = True,
    minimum_efficiency: float = 0.3,
) -> pd.DataFrame:

    filler_dic = df.groupby(by="fuel_draf").mean()["efficiency_estimate"].to_dict()
    df["filler"] = df["fuel_draf"].map(filler_dic)

    if fill_missing_efficiencies:
        number_of_nans = df["efficiency_estimate"].isna().sum()
        logger.info(f"{number_of_nans} nans in efficiencies of merit order filled")
        df["efficiency_estimate"].fillna(df["filler"], inplace=True)

    if fill_zscore_outlier:
        for fuel in df["fuel_draf"].unique():
            ser = df["efficiency_estimate"]
            ser = ser[ser.notnull()]
            cond_zscore = np.abs(stats.zscore(ser)) > zscore_threshold
            cond_fuel = df["fuel_draf"] == fuel
            number_of_nans = len(df[cond_zscore & cond_fuel])
            logger.info("f{number_of_nans} nans filled in fuel {fuel}.")
            df.loc[(cond_zscore & cond_fuel), "efficiency_estimate"] = filler_dic[fuel]

    df["eta_k"] = df["fuel_draf"].map(from_other.get_baumgaertner_data()["eta_k"])
    df["used_eff"] = df["efficiency_estimate"] if efficiency_per_plant else df["eta_k"]

    if ensure_minimum_efficiency:
        df.loc[df["used_eff"] < minimum_efficiency, "used_eff"] = minimum_efficiency
        logger.info(f"Minimum_efficiency set to {minimum_efficiency}.")

    return df


def _add_marginal_emissions_for_gen(df: pd.DataFrame, emission_data_source: str) -> pd.DataFrame:

    if emission_data_source == "baumgaertner":
        DATA_BAUMG = from_other.get_baumgaertner_data()
        df["mu_co2_k"] = df["fuel_draf"].map(DATA_BAUMG["mu_co2_k"])
        df["H_u_k"] = df["fuel_draf"].map(DATA_BAUMG["H_u_k"])
        df["marginal_emissions_for_gen"] = df["mu_co2_k"] / (df["H_u_k"] * df["used_eff"])

    elif emission_data_source == "quaschning":
        # in the quaschning data the mu_co2_k and the H_u_k are already included
        DATA_QUASCH = from_other.get_emissions_per_fuel_quaschning()
        df["marginal_emissions_for_gen"] = df["fuel_draf"].map(DATA_QUASCH) / df["used_eff"]

    else:
        raise ValueError("Emission_data_source must be either 'baumgaertner' or 'quaschning'.")

    return df


def _warn_if_not_enough_capa(mo_P: pd.DataFrame, resi_ser: pd.DataFrame) -> None:
    mo_max = mo_P["cumsum_capa"].max() / 1e3
    min_resi = resi_ser.min() / 1e3
    if mo_max < min_resi:
        logger.warning(
            f"Generation capacity ({mo_max:.2f} GW) is lower than "
            f"minimum residual load ({min_resi:.2f} GW)."
        )



import logging
from io import StringIO
from typing import Any, Callable, Dict, Iterable, List, Optional, Tuple, Union

import numpy as np
import pandas as pd

from elmada import get_mode
from elmada import helper as hp
from elmada import mappings as mp
from elmada import paths

logger = logging.getLogger(__name__)
logger.setLevel(logging.WARN)


def get_ETS_price(year: int) -> float:
    return get_ETS_prices()[year]


def get_ETS_prices() -> Dict:
    """Returns carbon emission prices of the EU emissions trading system (ETS) for different years.

    Source:
        * EEX (https://www.eex.com/en/market-data/environmental-markets/auction-market/european-emission-allowances-auction)
        via Sandbag (https://sandbag.org.uk/carbon-price-viewer/)
        * ICE via Quandl
    """
    mode = get_mode()
    if mode == "live":
        return get_ice_eua_prices()

    elif mode == "safe":
        return get_sandbag_eua_prices()
    else:
        raise RuntimeError()


def get_ice_eua_prices(cache: bool = True) -> Dict:
    """Get EUA prices from ICE [1] via Quandl [2].

    Sources:
        [1] https://www.theice.com/
        [2] https://www.quandl.com/data/CHRIS/ICE_C1-ECX-EUA-Futures-Continuous-Contract-1-C1-Front-Month
    """
    fp = paths.mode_dependent_cache_dir() / "QUANDL_eua_prices.parquet"

    if cache and fp.exists():
        ser = hp.read(fp)
    else:
        import quandl

        quandl.ApiConfig.api_key = hp.get_api_key("quandl")
        df = quandl.get("CHRIS/ICE_C1")
        ser = df["Settle"].resample("y").mean().rename("Price")
        ser.index = ser.index.year
        if cache:
            hp.write(ser, fp)

    return ser.to_dict()


def get_sandbag_eua_prices() -> Dict:
    """Get ETS EUA prices via Sandbag"""
    fp = paths.DATA_DIR / "sandbag/eua-price.csv"

    # # fix bad csv-syntax
    s = fp.read_text(encoding="utf8").replace('",', '";')

    io = StringIO(s)
    df = pd.read_csv(io, sep=";", decimal=",", index_col=0, skiprows=2)

    df.index = pd.to_datetime(df.index)
    df = df.resample("Y").mean()
    df = df.set_index(df.index.year, drop=True)
    return df.squeeze().to_dict()


def _get_light_oil_conversion_factor(source: int = 2):
    """Get the calorific value for light oil in MWh/Hektoliter from different sources.

    Sources:
    [1]: Heizöl Total (https://www.heizoel.total.de/rund-um-heizoel/aktuelles-tipps/heizoelkauf-beratung/heizwert-und-brennwert-von-heizol/)
        raw value: 9.8 kWh/l

    [2]: Mineralöl Franken (https://www.mineraloel-franken.de/fileadmin/redaktion/downloads/Produktspezifikation_Heizoel_EL_Total.pdf)
        raw values: 42.6 MJ/kg for light oil with a density of 860 kg/m³.
    """
    if source == 1:
        return 0.980

    if source == 2:
        return (
            42.6  # MJ/kg
            * 860  # kg/m³ ==> MJ/m³
            / 3600  # MJ/MWh ==> MWh/m³
            / 10  # Hektoliter/m³ ==> MWh/Hektoliter
        )


def _get_light_oil_price(year: int = 2019):
    """Get light oil price in €/MWh.

    Source: StatistischeBundesamt.2020 (https://www.destatis.de/DE/Themen/Wirtschaft/Preise/Publikationen/Energiepreise/energiepreisentwicklung-pdf-5619001.html)
    """
    data = pd.Series(
        {
            2005: 42.42,
            2006: 47.58,
            2007: 46.83,
            2008: 61.76,
            2009: 40.81,
            2010: 52.31,
            2011: 66.51,
            2012: 72.94,
            2013: 67.96,
            2014: 61.88,
            2015: 46.19,
            2016: 38.40,
            2017: 45.05,
            2018: 55.27,
            2019: 53.69,
        }
    )
    return (
        data[year] / _get_light_oil_conversion_factor()  # €/Hektoliter  # MWh/Hektoliter ==> €/MWh
    )


def _get_lignite_price(year: int = 2019, base_price: float = 6.3) -> float:
    """Get lignite price in €/MWh. Calculated as product of a destatis index and an baseprice for
    the year 2015.

    Source:
        indices: StatistischeBundesamt.2020 (https://www.destatis.de/DE/Themen/Wirtschaft/Preise/Publikationen/Energiepreise/energiepreisentwicklung-pdf-5619001.html)
        base_price: Konstantin.2017 (https://doi.org/10.1007/978-3-662-49823-1)
    """
    fp = paths.DATA_DIR / "destatis/energiepreisentwicklung-xlsx-5619001.xls"
    xl = pd.ExcelFile(fp)
    df = xl.parse(
        sheet_name=7, skiprows=28, header=None, skipfooter=0, index_col=0, na_values="-"
    ).dropna(axis=0, how="all")[13]

    df.index = df.index.str.slice(0, 5).astype(int)
    return df[year] * base_price / 100


def _get_coal_price(year: int = 2019, base_price: float = 10.12) -> float:
    """Get coal price in €/MWh. Calculated as product of a destatis index and an baseprice for
    the year 2015.

    Source:
        indices: StatistischeBundesamt.2020 (https://www.destatis.de/DE/Themen/Wirtschaft/Preise/Publikationen/Energiepreise/energiepreisentwicklung-pdf-5619001.html)
        base_price: Konstantin.2017 (https://doi.org/10.1007/978-3-662-49823-1)
    """
    fp = paths.DATA_DIR / "destatis/energiepreisentwicklung-xlsx-5619001.xls"
    xl = pd.ExcelFile(fp)
    df = xl.parse(
        sheet_name=7, skiprows=7, header=None, skipfooter=21, index_col=0, na_values="-"
    ).dropna(axis=0, how="all")[13]

    df.index = df.index.str.slice(0, 5).astype(int)
    return df[year] * base_price / 100


def _get_gas_price(year: int = 2019, country: str = "DE") -> float:
    """Get gas price in €/MWh. including excise taxes, without value-added tax.
    Data for most European countries in the years 2008 - 2019.

    Source: StatistischeBundesamt.2020 (https://www.destatis.de/DE/Themen/Wirtschaft/Preise/Publikationen/Energiepreise/energiepreisentwicklung-pdf-5619001.html)
    """
    fp = paths.DATA_DIR / "destatis/energiepreisentwicklung-xlsx-5619001.xls"
    xl = pd.ExcelFile(fp)
    nblocks = 4
    blocksize = 26
    block_list = []
    for i in range(nblocks):
        block_list.append(
            xl.parse(
                sheet_name=11,
                skiprows=4 + i * blocksize,
                skipfooter=blocksize * (nblocks - 1 - i) + 2,
                index_col=0,
                na_values="-",
            ).dropna(axis=0, how="all")
        )
    df = pd.concat(block_list, sort=False, join="outer", axis=1).dropna(axis=1, how="all")
    df["year"] = df.index.str.slice(5).astype(int)
    df = df.groupby("year").mean()  # get mean between year-halfs

    df = df / 100 * 1000  # convert cent/kWh into €/MWh

    default_value = df.mean().mean()
    warn_message = f"No data for gas price for {year},{country} ==> Default value {default_value:.2f} €/MWh is given."
    try:
        price = df.loc[year, mp.EU_de_for_gas_price[country]]
    except KeyError:
        logger.warning(warn_message)
        return default_value

    if np.isnan(price):
        logger.warning(warn_message)
        return default_value

    else:
        return price


def get_fuel_prices(year: int = 2019, country: str = "DE") -> Dict:
    """Get x_k: Fuel price [€ / MWh]

    Source:
        Nuclear: Konstantin.2017 (https://doi.org/10.1007/978-3-662-49823-1)
        Rest:
            StatistischeBundesamt.2020 (https://www.destatis.de/DE/Themen/Wirtschaft/Preise/Publikationen/Energiepreise/energiepreisentwicklung-pdf-5619001.html)
            + Konstantin.2017 (https://doi.org/10.1007/978-3-662-49823-1)
    """
    gas_price = _get_gas_price(year=year, country=country)
    try:
        return dict(
            oil=_get_light_oil_price(year=year),
            gas_cc=gas_price,
            gas=gas_price,
            coal=_get_coal_price(year=year),
            lignite=_get_lignite_price(year=year),
            nuclear=4.18,
        )
    except KeyError:  # if values for future years are missing
        return get_fuel_prices(year=2019)


def get_baumgaertner_data() -> Dict:
    """Nomentclature

    Per Fuel-type k
        - x_k: Fuel price [€ / MWh]
        - mu_co2_k: Carbon emission intensity [t_CO2eq / t_k]
        - H_u_k: Calorific value [MWh / t_k]
        - eta_k: Efficiency [MWh_el / MWh]

    Source: Baumgärtner.2019 (https://doi.org/10.1016/j.apenergy.2019.04.029)
        fuel prices: Konstantin.2017 (https://doi.org/10.1007/978-3-662-49823-1)
    """
    return dict(
        x_k=dict(oil=38.63, gas_cc=26.71, gas=29.81, coal=10.12, lignite=6.3, nuclear=4.18),
        mu_co2_k=dict(oil=3.15, gas_cc=1.93, gas=1.93, coal=2.6, lignite=1.39, nuclear=0.0),
        H_u_k=dict(
            oil=42.6 / 3.6,
            gas_cc=38.27 / 3.6,
            gas=38.27 / 3.6,
            coal=26.47 / 3.6,
            lignite=14.99 / 3.6,
            nuclear=41667.0,
        ),
        eta_k=dict(oil=0.373, gas_cc=0.50, gas=0.349, coal=0.377, lignite=0.356, nuclear=0.32),
    )


def get_emissions_per_fuel_quaschning() -> Dict:
    """Carbon emissions per fuel type [t_CO2eq / MWh]

    Source: Quaschning.2015 (https://www.volker-quaschning.de/datserv/CO2-spez/index_e.php)"""
    return dict(oil=0.28, gas_cc=0.25, gas=0.25, coal=0.34, lignite=0.36, nuclear=0.0)


def prepare_transmission_losses() -> pd.DataFrame:
    """Source: WorldBank.2020 (https://databank.worldbank.org/reports.aspx?source=2&series=EG.ELC.LOSS.ZS)"""
    fp = paths.DATA_DIR / "worldbank/Data_Extract_From_World_Development_Indicators/Data.csv"
    df = pd.read_csv(fp, nrows=58, na_values="..", index_col=2).iloc[:, 3:]
    df = df.rename(columns={k: k[:4] for k in df.keys()})
    df["mean"] = df.loc[:, "2010":"2014"].mean(1)
    return df


def get_transmission_efficiency_series(fillna: bool = True, cache: bool = True) -> pd.Series:
    """Returns a Series with transmission efficiencies and long country names in index."""

    fp = paths.mode_dependent_cache_dir() / "transmission_efficiencies.parquet"

    if cache and fp.exists():
        ser = hp.read(fp)

    else:
        df = prepare_transmission_losses()
        ser = df["mean"]
        if cache:
            hp.write(ser, fp)

    if fillna:
        na_countries = list(ser[ser.isna()].index)
        filler = ser.mean()
        ser = ser.fillna(filler)
        logger.info(f"No data for transmission efficiency for {na_countries}")

    return 1 - (ser / 100)


def get_transmission_efficiency(country: str) -> float:
    """Returns scalar value of transmission efficiency."""
    ser = get_transmission_efficiency_series()

    try:
        cy_long = mp.EUROPE_COUNTRIES[country]
        return ser[cy_long]

    except KeyError:
        default = ser.mean()
        logger.warning(
            f"No transmission efficiency data for {country}. European mean of {default} given."
        )
        return default



class ConfigUtil:
    mode = "safe"


def set_mode(mode: str):
    assert mode in ["safe", "live"]
    ConfigUtil.mode = mode


def get_mode() -> str:
    return ConfigUtil.mode


def is_safe_mode() -> bool:
    return get_mode() == "safe"



from pathlib import Path

from setuptools import find_packages, setup

exec(Path("elmada/_version.py").read_text().strip())  # Set the __version__ variable

long_description = Path("README.md").read_text().strip()

setup(
    name="elmada",
    version=__version__,
    author="Markus Fleschutz",
    author_email="mfleschutz@gmail.com",
    description="Electricity market data",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/DrafProject/elmada",
    license="LGPLv3",
    packages=find_packages(exclude=["doc", "tests"]),
    python_requires=">=3.7",
    install_requires=[
        "appdirs",
        "bs4",
        "entsoe-py==0.2.10",
        "ipython",
        "lxml",
        "matplotlib",
        "numpy",
        "pandas",
        "pyarrow",
        "quandl",
        "requests",
        "scipy",
        "xlrd",
    ],
    extras_require={"dev": ["plotly", "pytest", "pytest-cov", "pytest-mock", "mypy"]},
    include_package_data=True,
    package_data={"elmada": ["*.parquet", "*.csv", "*.txt", "*xls"]},
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Science/Research",
        "Intended Audience :: Developers",
        "Topic :: Scientific/Engineering",
        "License :: OSI Approved :: GNU Lesser General Public License v3 (LGPLv3)",
        "Natural Language :: English",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3 :: Only",
    ],
    keywords=["energy market data", "energy systems", "carbon emission factors", "demand response"],
)



"""Various mappings including countries, fuel types, colors.

ISO 3166-1 alpha-2 country codes are used, see https://en.wikipedia.org/wiki/ISO_3166-1

# TODO: there is a python module for ISO 3166-1, see https://github.com/deactivated/python-iso3166

"""

import collections
from typing import Any, Callable, Dict, Iterable, List, Optional, Set, Tuple, Union

# EU 27
EU = {
    "AT": "Austria",
    "BE": "Belgium",
    "BG": "Bulgaria",
    "CY": "Cyprus",
    "CZ": "Czech Republic",
    "DE": "Germany",
    "DK": "Denmark",
    "EE": "Estonia",
    "ES": "Spain",
    "FI": "Finland",
    "FR": "France",
    "GR": "Greece",
    "HR": "Croatia",
    "HU": "Hungary",
    "IE": "Ireland",
    "IT": "Italy",
    "LT": "Lithuania",
    "LU": "Luxembourg",
    "LV": "Latvia",
    "MT": "Malta",
    "NL": "Netherlands",
    "PL": "Poland",
    "PT": "Portugal",
    "RO": "Romania",
    "SE": "Sweden",
    "SI": "Slovenia",
    "SK": "Slovakia",
}

# Non-EU27 European countries
OTHER = {
    "AD": "Andorra",
    "AL": "Albania",
    "AM": "Armenia",
    "AZ": "Azerbaijan",
    "BA": "Bosnia and Herzegovina",
    "BY": "Belarus",
    "CH": "Switzerland",
    "FO": "Faeroe Islands",
    "GB": "United Kingdom",
    "GE": "Georgia",
    "GI": "Gibraltar",
    "IS": "Iceland",
    "KZ": "Kazakhstan",
    "LI": "Liechtenstein",
    "MC": "Monaco",
    "MD": "Moldova",
    "ME": "Montenegro",
    "MK": "Macedonia",
    "NO": "Norway",
    "RS": "Serbia",
    "RU": "Russian Federation",
    "SM": "San Marino",
    "TR": "Turkey",
    "UA": "Ukraine",
    "VA": "Vatican City State",
    "XK": "Kosovo",
}

# Countries that are excluded form the analysis
EXCLUDED = {
    "AD": "No generation data available from ENTSO-E",
    "AM": "No generation data available from ENTSO-E",
    "AZ": "No generation data available from ENTSO-E",
    "CY": "No generation data available from ENTSO-E",
    "FO": "No generation data available from ENTSO-E",
    "GE": "No generation data available from ENTSO-E",
    "GI": "No generation data available from ENTSO-E",
    "IS": "No generation data available from ENTSO-E",
    "KZ": "No generation data available from ENTSO-E",
    "LI": "No generation data available from ENTSO-E",
    "MC": "No generation data available from ENTSO-E",
    "MD": "No generation data available from ENTSO-E",
    "SM": "No generation data available from ENTSO-E",
    "VA": "No generation data available from ENTSO-E",
    "XK": "No generation data available from ENTSO-E",
    "LU": "No generation data available from ENTSO-E",
    "RU": "No generation data available from ENTSO-E",
    "TR": "No generation data available from ENTSO-E",
    "MT": "No generation data available from ENTSO-E",
    "HR": "No generation data available from ENTSO-E",
    "UA": "No generation data available from ENTSO-E",
    "AL": "No generation data available from ENTSO-E",
    "BY": "No generation data available from ENTSO-E",
    "BA": "Only coal as conventional generation data from ENTSO-E.",
    "CH": "Only nuclear as conventional generation data from ENTSO-E.",
    "SE": "Only nuclear as conventional generation data from ENTSO-E.",
    "NO": "Only gas as conventional generation data from ENTSO-E.",
    "LV": "Only gas as conventional generation data from ENTSO-E.",
    "MK": "Only lignite as conventional generation data from ENTSO-E.",
    "ME": "Only lignite as conventional generation data from ENTSO-E.",
    "EE": "Only other_conv as conventional generation data from ENTSO-E.",
    # "LT": "Only gas & oil as conventional generation data from ENTSO-E.",
    "BG": "No installed capacity for 2019 from ENTSO-E",
    "SK": "No installed capacity for 2019 from ENTSO-E",
}

EUROPE_COUNTRIES = {**EU, **OTHER}

EUROPE_COUNTRIES_LONG_TO_SHORT = {long: short for short, long in EUROPE_COUNTRIES.items()}

d = {k: v for k, v in EUROPE_COUNTRIES.items() if k not in EXCLUDED}
COUNTRIES_FOR_ANALYSIS = collections.OrderedDict(sorted(d.items()))

# Modified entsoe-py mapping to match whole countries
NEIGHBOURS_CY = {
    "AT": ["CH", "CZ", "DE", "HU", "IT", "SI"],
    "BA": ["HR", "ME", "RS"],
    "BE": ["NL", "DE", "FR", "GB"],
    "BG": ["GR", "MK", "RO", "RS", "TR"],
    "CH": ["AT", "DE", "FR", "IT"],
    "CZ": ["AT", "DE", "PL", "SK"],
    "DE": ["AT", "BE", "CH", "CZ", "DK", "FR", "IT", "NL", "PL", "SE", "SI"],
    "DK": ["DE", "NO", "SE"],
    "EE": ["FI", "LV", "RU"],
    "ES": ["FR", "PT"],
    "FI": ["EE", "NO", "RU", "SE", "SE"],
    "FR": ["BE", "CH", "DE", "ES", "GB", "IT"],
    "GB": ["BE", "FR", "IE", "NL"],
    "GR": ["AL", "BG", "IT", "MK", "TR"],
    "HU": ["AT", "HR", "RO", "RS", "SK", "UA"],
    "IE": ["GB"],
    "IT": ["AT", "FR", "DE", "GR", "MT", "ME", "SI", "CH"],
    "LT": ["BY", "LV", "PL", "RU", "SE"],
    "LV": ["EE", "LT", "RU"],
    "ME": ["AL", "BA", "RS"],
    "MK": ["BG", "GR", "RS"],
    "MT": ["IT"],
    "NL": ["BE", "DE", "DE", "GB", "NO"],
    "NO": ["SE", "FI", "DK", "NL"],
    "PL": ["CZ", "DE", "DE", "LT", "SE", "SK", "UA"],
    "PT": ["ES"],
    "RO": ["BG", "HU", "RS", "UA"],
    "RS": ["AL", "BA", "BG", "HR", "HU", "ME", "MK", "RO"],
    "SE": ["FI", "NO", "DE", "DK", "LT", "PL"],
    "SI": ["AT", "DE", "HR", "IT"],
    "SK": ["CZ", "HU", "PL", "UA"],
}


EU_de_for_gas_price = {
    "AT": "Österreich",
    "BE": "Belgien",
    "BG": "Bulgarien",
    "CZ": "Tschechiensche\nRepublik",
    "DE": "Deutschland",
    "DK": "Dänemark",
    "EE": "Estland",
    "ES": "Spanien",
    "FI": "Finnland",
    "FR": "Frankreich",
    "GB": "Vereinigtes\nKönigreich",
    "GE": "Georgien",
    "GR": "Griechenland",
    "HR": "Kroatien",
    "HU": "Ungarn",
    "IE": "Irland",
    "IT": "Italien",
    "LT": "Litauen",
    "LU": "Luxemburg",
    "LV": "Lettland",
    "MK": "Mazedonien",
    "NL": "Niederlande",
    "PO": "Polen",
    "PT": "Portugal",
    "RO": "Rumänien",
    "SE": "Schweden",
    "SI": "Slowenien",
    "SK": "Slowakei",
    "TR": "Türkei",
    "UA": "Ukraine",
}

TECH_COLORS = {
    "load": "grey",
    "biomass": "olive",
    "hydro": "lightseagreen",
    "solar": "gold",
    "wind_onshore": "royalblue",
    "wind_offshore": "navy",
    "other_RES": "lightgreen",
    "nuclear": "deeppink",
    "lignite": "chocolate",
    "coal": "saddlebrown",
    "pumped_hydro": "mediumseagreen",
    "gas": "darkgreen",
    "gas_cc": "lightgreen",
    "oil": "darkgrey",
    "other_conv": "lightgray",
    "nan": "red",
}

ORDERED_ENTSOE_FUELS = [
    "Solar",
    "Wind Offshore",
    "Wind Onshore",
    "Other renewable",
    "Geothermal",
    "Waste",
    "Biomass",
    "Hydro Run-of-river and poundage",
    "Hydro Water Reservoir",
    "Marine",
    "Nuclear",
    "Fossil Brown coal/Lignite",
    "Fossil Hard coal",
    "Fossil Gas",
    "Fossil Oil",
    "Other_conventional",
    "Other",
    "Fossil Oil shale",
    "Fossil Coal-derived gas",
    "Fossil Peat",
    "Hydro Pumped Storage",
]


FUEL_AGGREGATION = {
    "Other": "other_conv",
    "other_conventional": "other_conv",
    "Fossil Coal-derived gas": "other_conv",
    "Fossil Oil shale": "other_conv",
    "Fossil Peat": "other_conv",
    "Waste": "other_RES",
    "Geothermal": "other_RES",
    "Hydro Water Reservoir": "other_RES",
    "Marine": "other_RES",
    "Other renewable": "other_RES",
    "other renewables": "other_RES",
}

FUEL_RENAME = {
    "Biomass": "biomass",
    "Fossil Brown coal/Lignite": "lignite",
    "Fossil Oil": "oil",
    "Fossil Gas": "gas",
    "Fossil Hard coal": "coal",
    "Hydro Pumped Storage": "pumped_hydro",
    "Hydro Run-of-river and poundage": "hydro",
    "Nuclear": "nuclear",
    "Solar": "solar",
    "Wind Offshore": "wind_offshore",
    "Wind Onshore": "wind_onshore",
}

OPSD_TO_DRAF = {
    "Hard coal": "coal",
    "Lignite": "lignite",
    "Natural gas": "gas",
    "Nuclear": "nuclear",
    "Oil": "oil",
}

# Fuels after aggregation
RES = ["hydro", "biomass", "solar", "wind_offshore", "wind_onshore", "other_RES"]
CONV = ["nuclear", "lignite", "coal", "pumped_hydro", "gas", "oil", "other_conv"]
DRAF_FUELS = RES + CONV

# Fuels for calculation of residual load (simplified merit order simulation)
CONV2 = ["nuclear", "lignite", "coal", "gas", "oil"]
RES2 = [f for f in DRAF_FUELS if f not in CONV2]

# Fuels for simplified merit oder simulation
PWL_FUELS = ["nuclear", "lignite", "coal", "gas_cc", "gas", "oil"]


def get_tech_colors(technologies: Iterable[str]) -> List[str]:
    """Returns list of colours for given list of technologies."""
    return [TECH_COLORS[i] for i in technologies]



from ._version import __version__

# isort: off

from .mode import get_mode, set_mode
from .helper import set_api_keys, make_symlink_to_cache

# isort: on

from . import (
    cc_share,
    eu_pwl,
    from_entsoe,
    from_geo_scraped,
    from_geo_via_morph,
    from_opsd,
    from_other,
    from_smard,
    helper,
    paths,
    plots,
)
from .main import (
    get_el_national_generation,
    get_emissions,
    get_merit_order,
    get_prices,
    get_residual_load,
)



from typing import Any, Callable, Dict, Iterable, List, Optional, Tuple, Union

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from matplotlib.patches import Patch

from elmada import helper as hp
from elmada import mappings as mp
from elmada.main import get_emissions, get_merit_order, get_residual_load


def merit_order(
    year=2019,
    country="DE",
    method="PWL",
    ylim_left: Optional[float] = None,
    ylim_right: Optional[float] = None,
    ax: Optional["plt.axes"] = None,
    mo_P: Optional[pd.DataFrame] = None,
    include_histo: bool = False,
    include_legend: bool = True,
    legend_fontsize: float = 9,
    **mo_kwargs,
) -> Tuple[plt.Figure, plt.Axes, plt.Axes]:
    """Plot the merit order with the given method."""

    plt.rcParams.update({"mathtext.default": "regular"})
    plt.rc("text", usetex=False)
    plt.rc("font", family="serif")

    if mo_P is None:
        mo_P = get_merit_order(year=year, country=country, method=method, **mo_kwargs)

    mo_P = mo_P.copy()  # avoids modifying lru_cache
    # the unit of mo_P.marginal_emissions is kg
    mo_P.cumsum_capa /= 1000  # from MW to GW
    mo_P.capa /= 1000  # from MW to GW

    if ax is not None:
        fig = ax.get_figure()
    else:
        fig, ax = plt.subplots(1, figsize=(7, 3))

    ax_right = ax.twinx()

    color_left = "darkblue"
    color_right = "black"

    ax.step(
        x=hp.add_row_for_steps(mo_P.cumsum_capa),
        y=hp.add_row_for_steps(mo_P.marginal_cost),
        where="pre",
        label=r"pwl $c_p^m$",
        drawstyle="steps",
        color=color_left,
        linewidth=1.5,
    )

    ax_right.step(
        x=hp.add_row_for_steps(mo_P.cumsum_capa),
        y=hp.add_row_for_steps(mo_P.marginal_emissions),
        where="pre",
        label=r"pwl $\varepsilon_p$",
        drawstyle="steps",
        color=color_right,
        linewidth=0.5,
    )

    ax.set_xlabel(r"Net power [GW$_{el}$]")
    ax.set_ylabel(r"Marginal cost $c_p^m$" + "\n" + r"[€/$MWh_{el}$]")
    ax_right.set_ylabel(
        r"Marginal emissions $\varepsilon_p$" + "\n" + r"[$t_{CO_{2}eq}$/$MWh_{el}$]"
    )

    plt.xlim((0, mo_P.cumsum_capa.max()))
    ax.set_ylim(bottom=0, top=ylim_left)
    ax_right.set_ylim(bottom=0, top=ylim_right)

    ax.yaxis.label.set_color(color_left)
    ax.tick_params(axis="y", colors=color_left)
    ax.spines["left"].set_color(color_left)

    ax_right.yaxis.label.set_color(color_right)
    ax_right.tick_params(axis="y", colors=color_right)
    ax_right.spines["right"].set_color(color_right)

    alpha_value_for_fuels = _make_colored_background_for_fuel_types(ax, mo_P)

    if include_legend:
        _add_legend(mo_P, legend_fontsize, alpha_value_for_fuels)

    if include_histo:
        _add_histo(year, country, ax_right)

    plt.tight_layout()

    return fig, ax, ax_right


def _add_histo(year, country, ax_right):
    resi = get_residual_load(year=year, country=country) / 1000
    n, bins = np.histogram(resi, bins=20, density=True)
    ylim = ax_right.get_ylim()[1]
    max_loc = 0.15
    y = n * max_loc * (ylim / n.max())
    ax_right.fill_between(bins[1:], y, step="post", color="black", alpha=0.5)


def _add_legend(mo_P, legend_fontsize, alpha_value_for_fuels):
    fuels = mo_P.fuel_draf.unique()
    cdict = mp.TECH_COLORS
    legend_elements = [
        Patch(facecolor=cdict[f], edgecolor=cdict[f], alpha=alpha_value_for_fuels, label=f)
        for f in fuels
    ]
    leg = plt.legend(
        handles=legend_elements,
        bbox_to_anchor=(0.0, 1.01, 1.0, 0.102),
        loc="lower left",
        ncol=6,
        mode="expand",
        borderaxespad=0.0,
        fontsize=legend_fontsize,
        framealpha=1.0,
        handletextpad=0.5,
        edgecolor="inherit",
    )
    leg.get_frame().set_linewidth(0.5)


def _make_colored_background_for_fuel_types(ax, mo_P):
    alpha_value_for_fuels = 0.7
    for i, row in mo_P.iterrows():
        ax.axvspan(
            row["cumsum_capa"] - row["capa"],
            row["cumsum_capa"],
            facecolor=mp.TECH_COLORS[row["fuel_draf"]],
            linewidth=0,
            alpha=alpha_value_for_fuels,
            ymin=0,
            ymax=1,
        )
    return alpha_value_for_fuels


def cefs_scatter_plotly(
    year: int = 2019, freq: str = "60min", country: str = "DE", method: str = "MEF_PWL", **mo_kwargs
):
    """Optional interactive scatter plot. Works only if plotly is installed."""
    import plotly.express as px

    first_part, second_part = method.split("_")

    df = get_emissions(year=year, freq=freq, country=country, method=f"_{second_part}", **mo_kwargs)
    return px.scatter(
        df, y=f"{first_part}s", color="marginal_fuel", color_discrete_map=mp.TECH_COLORS
    )


def cefs_scatter(
    year=2019,
    freq: str = "60min",
    country: str = "DE",
    method: str = "MEF_PWL",
    figsize: Tuple = (8, 3),
    **mo_kwargs,
) -> None:
    """Plot the CEFs as scatter plot with colors refering to marginal power plant type."""
    _, second_part = method.split("_")

    df = get_emissions(year=year, freq=freq, country=country, method=f"_{second_part}", **mo_kwargs)

    df = df.reset_index().set_index(["index", "marginal_fuel"])["MEFs"].unstack()
    df.index.name = None
    df.plot(
        marker="o",
        markersize=2.5,
        figsize=figsize,
        linewidth=0,
        color=mp.get_tech_colors(df.keys()),
    )

    ax = plt.gca()
    ax.set(ylabel="Carbon emission factors\n[$g_{CO_{2}eq}$ / $kWh_{el}$]")
    ax.legend(ncol=6, loc="lower center", bbox_to_anchor=(0.5, 1), frameon=True)
